import pandas as pd
import numpy as np
import shap

from shapash.backend.base_backend import BaseBackend


class ShapBackend(BaseBackend):
    # When grouping features contributions together, Shap uses the sum of the contributions
    # of the features that belong to the group
    column_aggregation = 'sum'
    name = 'shap'

    def __init__(self, model, preprocessing=None, explainer_args=None, explainer_compute_args=None):
        super(ShapBackend, self).__init__(model, preprocessing)
        self.explainer_args = explainer_args if explainer_args else {}
        self.explainer_compute_args = explainer_compute_args if explainer_compute_args else {}
        self.explainer = shap.Explainer(model=model, **self.explainer_args)

    def run_explainer(self, x: pd.DataFrame) -> dict:
        """
        Computes and returns local contributions using Shap explainer

        Parameters
        ----------
        x : pd.DataFrame
            The observations dataframe used by the model

        Returns
        -------
        explain_data : pd.DataFrame or list of pd.DataFrame
            local contributions
        """
        contributions = self.explainer(x, **self.explainer_compute_args)
        explain_data = dict(contributions=contributions.values)
        return explain_data


def get_shap_interaction_values(x_df, explainer):
    """
    Compute the shap interaction values for a given dataframe.
    Also checks if the explainer is a TreeExplainer.

    Parameters
    ----------
    x_df : pd.DataFrame
        DataFrame for which will be computed the interaction values using the explainer.
    explainer : shap.TreeExplainer
        explainer object used to compute the interaction values.

    Returns
    -------
    shap_interaction_values : np.ndarray
        Shap interaction values for each sample as an array of shape (# samples x # features x # features).
    """
    if not isinstance(explainer, shap.TreeExplainer):
        raise ValueError(f"Explainer type ({type(explainer)}) is not a TreeExplainer. "
                         f"Shap interaction values can only be computed for TreeExplainer types")

    shap_interaction_values = explainer.shap_interaction_values(x_df)

    # For models with vector outputs the previous function returns one array for each output.
    # We sum the contributions here.
    if isinstance(shap_interaction_values, list):
        shap_interaction_values = np.sum(shap_interaction_values, axis=0)

    return shap_interaction_values



from typing import Any, Optional, List, Union

import numpy as np
import pandas as pd

from shapash.backend.base_backend import BaseBackend
from shapash.utils.transform import get_preprocessing_mapping

try:
    from acv_explainers import ACVTree
    from acv_explainers.utils import get_null_coalition
    _is_acv_available = True
except ImportError:
    _is_acv_available = False


class AcvBackend(BaseBackend):
    # Coalitions should be grouped using one column value only and not the sum like shap
    column_aggregation = 'first'
    name = 'acv'
    supported_cases = ['classification']

    def __init__(
            self,
            model,
            data=None,
            preprocessing=None,
            active_sdp=True,
            explainer_args=None,
            explainer_compute_args=None
    ):
        if _is_acv_available is False:
            raise ValueError(
                """
                Active Shapley values requires the ACV package,
                which can be installed using 'pip install acv-exp'
                """
            )
        super(AcvBackend, self).__init__(model, preprocessing)
        self.active_sdp = active_sdp
        self.data = data
        self.explainer_args = explainer_args if explainer_args else {}
        self.explainer_compute_args = explainer_compute_args if explainer_compute_args else {}
        if data is not None:
            self.explainer = ACVTree(model=model, data=data, **self.explainer_args)
        else:
            self.explainer = None

    def run_explainer(self, x: pd.DataFrame) -> dict:
        if self.data is None:
            # This is used to handle the case where data object was not definied 
            self.data = x
            self.explainer = ACVTree(model=self.model, data=self.data, **self.explainer_args)

        explain_data = {}

        mapping = get_preprocessing_mapping(x, self.preprocessing)
        c = []
        for col in mapping.keys():
            if len(mapping[col]) > 1:
                c.append([x.columns.to_list().index(col_i) for col_i in mapping[col]])
        if len(c) == 0:
            c = [[]]

        sdp_importance, sdp_index, size, sdp = self.explainer.importance_sdp_clf(
            X=x.values,
            data=np.asarray(self.data)
        )
        s_star, n_star = get_null_coalition(sdp_index, size)
        contributions = self.explainer.shap_values_acv_adap(
            X=x.values,
            C=c,
            S_star=s_star,
            N_star=n_star,
            size=size
        )
        if contributions.shape[-1] > 1:
            contributions = [pd.DataFrame(contributions[:, :, i], columns=x.columns, index=x.index)
                             for i in range(contributions.shape[-1])]
        else:
            contributions = pd.DataFrame(contributions[:, :, 0], columns=x.columns, index=x.index)

        explain_data['sdp'] = sdp
        explain_data['sdp_index'] = sdp_index
        explain_data['init_columns'] = x.columns.to_list()
        explain_data['contributions'] = contributions
        explain_data['features_mapping'] = mapping

        return explain_data

    def get_global_features_importance(
            self,
            contributions: Union[pd.DataFrame, List[pd.DataFrame]],
            explain_data: Any = None,
            subset: Optional[List[int]] = None
    ) -> Union[pd.Series, List[pd.Series]]:

        count_cols = {i: 0 for i in range(len(explain_data['sdp_index'][0]))}

        for i, list_imp_feat in enumerate(explain_data['sdp_index']):
            for col in list_imp_feat:
                if col != -1 and explain_data['sdp'][i] > 0.9:
                    count_cols[col] += 1

        features_cols = {explain_data['init_columns'][k]: v for k, v in count_cols.items()}

        mapping = explain_data['features_mapping']
        list_cols_ohe = [c for list_c in mapping.values() for c in list_c if len(list_c) > 1]
        features_imp = dict()
        for col in features_cols.keys():
            if col in list_cols_ohe:
                for col_mapping in mapping.keys():
                    if col in mapping[col_mapping]:
                        features_imp[col_mapping] = features_cols[col]
            else:
                features_imp[col] = features_cols[col]

        features_imp = pd.Series(features_imp).sort_values(ascending=True)
        if self._case == 'classification':
            features_imp = [features_imp for _ in range(len(contributions))]
        return features_imp




"""
Main class of Web application Shapash
"""
import dash
from dash import dash_table
import dash_daq as daq
from dash import no_update
from dash.exceptions import PreventUpdate
import dash_bootstrap_components as dbc
from dash import dcc
from dash import html
from dash import MATCH, ALL
from dash.dependencies import Output, Input, State
from flask import Flask
import pandas as pd
import plotly.graph_objs as go
import random
import numpy as np
import datetime
import re
from math import log10
from shapash.webapp.utils.utils import check_row, round_to_k
from shapash.webapp.utils.MyGraph import MyGraph
from shapash.utils.utils import truncate_str
from shapash.webapp.utils.explanations import Explanations


def create_input_modal(id, label, tooltip):
    return dbc.Row(
        [
            dbc.Label(label, id=f'{id}_label', html_for=id, width=8),
            dbc.Col(
                dbc.Input(id=id, type="number", value=0),
                width=4),
            dbc.Tooltip(tooltip, target=f'{id}_label', placement='bottom'),
        ], className="g-3"
    )


class SmartApp:
    """
        Bridge pattern decoupling the application part from SmartExplainer and SmartPlotter.
        Attributes
        ----------
        explainer: object
            SmartExplainer instance to point to.
    """

    def __init__(self, explainer, settings: dict = None):
        """
        Init on class instantiation, everything to be able to run the app on server.
        Parameters
        ----------
        explainer : SmartExplainer
            SmartExplainer object
        settings : dict
            A dict describing the default webapp settings values to be used
            Possible settings (dict keys) are 'rows', 'points', 'violin', 'features'
            Values should be positive ints
        """
        # APP
        self.server = Flask(__name__)
        self.app = dash.Dash(
            server=self.server,
            external_stylesheets=[dbc.themes.BOOTSTRAP],
        )
        self.app.title = 'Shapash Monitor'
        if explainer.title_story:
            self.app.title += ' - ' + explainer.title_story
        self.explainer = explainer

        # SETTINGS
        self.logo = self.app.get_asset_url('shapash-fond-fonce.png')
        self.color = explainer.plot._style_dict["webapp_button"]
        self.bkg_color = explainer.plot._style_dict["webapp_bkg"]
        self.title_menu_color = explainer.plot._style_dict["webapp_title"]
        self.settings_ini = {
            'rows': 1000,
            'points': 1000,
            'violin': 10,
            'features': 20,
        }
        if settings is not None:
            for k, v in self.settings_ini.items():
                self.settings_ini[k] = settings[k] if k in settings and isinstance(
                    settings[k], int) and 0 < settings[k] else v
        self.settings = self.settings_ini.copy()

        self.predict_col = ['_predict_']
        self.explainer.features_imp = self.explainer.state.compute_features_import(
            self.explainer.contributions)
        if self.explainer._case == 'classification':
            self.label = self.explainer.check_label_name(len(self.explainer._classes) - 1, 'num')[1]
            self.selected_feature = self.explainer.features_imp[-1].idxmax()
            self.max_threshold = int(max([x.applymap(lambda x: round_to_k(x, k=1)).max().max()
                                          for x in self.explainer.contributions]))
        else:
            self.label = None
            self.selected_feature = self.explainer.features_imp.idxmax()
            self.max_threshold = int(self.explainer.contributions.applymap(
                lambda x: round_to_k(x, k=1)).max().max())
        self.list_index = []
        self.subset = None
        self.last_click_data = None

        # DATA
        self.explanations = Explanations()  # To get explanations of "?" buttons
        self.dataframe = pd.DataFrame()
        self.round_dataframe = pd.DataFrame()
        self.init_data()

        # COMPONENTS
        self.components = {
            'menu': {},
            'table': {},
            'graph': {},
            'filter': {},
            'settings': {}
        }
        self.init_components()

        # LAYOUT
        self.skeleton = {
            'navbar': {},
            'body': {}
        }
        self.make_skeleton()
        self.app.layout = html.Div([self.skeleton['navbar'], self.skeleton['body']])

        # CALLBACK
        self.callback_fullscreen_buttons()
        self.init_callback_settings()
        self.callback_generator()

    def init_data(self):
        """
        Method which initializes data from explainer object
        """
        if hasattr(self.explainer, 'y_pred'):
            self.dataframe = self.explainer.x_init.copy()
            if isinstance(self.explainer.y_pred, (pd.Series, pd.DataFrame)):
                self.predict_col = self.explainer.y_pred.columns.to_list()[0]
                self.dataframe = self.dataframe.join(self.explainer.y_pred)
            elif isinstance(self.explainer.y_pred, list):
                self.dataframe = self.dataframe.join(
                    pd.DataFrame(data=self.explainer.y_pred,
                                 columns=[self.predict_col],
                                 index=self.explainer.x_init.index)
                    )
            else:
                raise TypeError('y_pred must be of type pd.Series, pd.DataFrame or list')
        else:
            raise ValueError('y_pred must be set when calling compile function.')

        self.dataframe['_index_'] = self.explainer.x_init.index
        self.dataframe.rename(columns={f'{self.predict_col}': '_predict_'}, inplace=True)
        col_order = ['_index_', '_predict_'] + self.dataframe.columns.drop(['_index_', '_predict_']).tolist()
        random.seed(79)
        self.list_index = \
            random.sample(
                population=self.dataframe.index.tolist(),
                k=min(self.settings['rows'], len(self.dataframe.index.tolist()))
            )
        self.dataframe = self.dataframe[col_order].loc[self.list_index].sort_index()
        self.round_dataframe = self.dataframe.copy()
        for col in list(self.dataframe.columns):
            typ = self.dataframe[col].dtype
            if typ == float:
                std = self.dataframe[col].std()
                if std != 0:
                    digit = max(round(log10(1 / std) + 1) + 2, 0)
                    self.round_dataframe[col] = \
                        self.dataframe[col].map(f'{{:.{digit}f}}'.format).astype(float)

    def init_components(self):
        """
        Initialize components (graph, table, filter, settings, ...) and insert it inside
        components containers which are created by init_skeleton
        """

        self.components['settings']['input_rows'] = create_input_modal(
            id='rows',
            label="Number of rows for subset",
            tooltip="Set max number of lines for subset (datatable). \
                     Filter will be apply on this subset."
        )

        self.components['settings']['input_points'] = create_input_modal(
            id='points',
            label="Number of points for plot",
            tooltip="Set max number of points in feature contribution plots."
        )

        self.components['settings']['input_features'] = create_input_modal(
            id='features',
            label="Number of features to plot",
            tooltip="Set max number of features to plot in features \
                     importance and local explanation plots."
        )

        self.components['settings']['input_violin'] = create_input_modal(
            id='violin',
            label="Max number of labels for violin plot",
            tooltip="Set max number of labels to display a violin plot \
                     for feature contribution plot (otherwise a scatter \
                                                    plot is displayed)."
        )

        self.components['settings']['name'] = dbc.Row(
            [
                dbc.Checklist(
                    options=[{"label": "Use domain name for \
                              features name.", "value": 1}], value=[], inline=True,
                    id="name",
                    style={"margin-left": "20px"}
                ),
                dbc.Tooltip("Replace technical feature names by \
                            domain names if exists.",
                            target='name', placement='bottom'),
            ], className="g-3",
        )

        self.components['settings']['modal'] = dbc.Modal(
            [
                dbc.ModalHeader("Settings"),
                dbc.ModalBody(
                    dbc.Form(
                        [
                            self.components['settings']['input_rows'],
                            self.components['settings']['input_points'],
                            self.components['settings']['input_features'],
                            self.components['settings']['input_violin'],
                            self.components['settings']['name']
                        ]
                    )
                ),
                dbc.ModalFooter(
                    dbc.Button("Apply", id="apply", className="ml-auto")
                ),
            ],
            id="modal"
        )

        self.components['menu'] = dbc.Row(
            [
                dbc.Col(
                    [
                        html.Div(
                            daq.BooleanSwitch(
                                id='bool_groups',
                                on=True,
                                style={'display': 'none'} if self.explainer.features_groups is None else {},
                                color=self.color[0],
                                label={
                                    'label': 'Groups',
                                    'style': {
                                        'fontSize': 18,
                                        'color': self.color[0],
                                        'fontWeight': 'bold',
                                        "margin-left": "5px"
                                    },
                                },
                                labelPosition="right"
                            ),
                            style={"margin-right": "35px"}
                        )
                    ],
                    width="auto", align="center",
                ),
                dbc.Col(
                    [
                        html.H4(
                            [dbc.Badge("Regression", id='regression_badge',
                                       style={"margin-right": "5px",
                                              "margin-left": "0px"},
                                       color=''),
                             dbc.Badge("Classification", id='classification_badge', color='')
                             ], style={"margin-right": "5px"}
                        ),
                    ],
                    width="auto", align="center", style={'padding': 'auto'}
                ),
                dbc.Col(
                    dbc.Collapse(
                        dbc.Row(
                            [
                                # 2 columns to have class beside the dropdown buttons
                                dbc.Col([
                                    dbc.Label("Class:", style={'color': 'white', 'margin': '0px'}),
                                ], align="center"),
                                dbc.Col([
                                    dcc.Dropdown(
                                        id="select_label",
                                        options=[], value=None,
                                        clearable=False, searchable=False,
                                        style={"verticalAlign": "middle",
                                               "zIndex": '1010',
                                               "min-width": '160px',
                                               'height': '100%'}
                                    )
                                ], style={"margin-right": "17px",
                                          "padding": "0px",
                                          "width": "auto"})
                            ],
                            style={"margin": "0px"}
                        ),
                        is_open=True, id='select_collapse'
                    ),
                    width="auto", align="center", style={'padding': '0px'}
                ),
                dbc.Col([
                    html.Div(
                        [
                            html.Img(id='settings', title='settings', alt='Settings',
                                     src=self.app.get_asset_url('settings.png'),
                                     height='40px',
                                     style={'cursor': 'pointer'}),
                            self.components['settings']['modal'],
                        ]
                    )],
                    align="center", width="auto", style={'padding': '0px'}
                )
            ],
            className="g-0", justify="end"
        )

        self.adjust_menu()

        self.components['table']['dataset'] = dash_table.DataTable(
            id='dataset',
            data=self.round_dataframe.to_dict('records'),
            tooltip_data=[
                {
                    column: {'value': str(value), 'type': 'text'}
                    for column, value in row.items()
                } for row in self.dataframe.to_dict('rows')
            ], tooltip_duration=2000,

            columns=[{"name": '_index_', "id": '_index_'},
                     {"name": '_predict_', "id": '_predict_'}] +
                    [{"name": i, "id": i} for i in self.explainer.x_init],
            editable=False, row_deletable=False,
            style_as_list_view=True,
            virtualization=True,
            page_action='none',
            fixed_rows={'headers': True, 'data': 0},
            fixed_columns={'headers': True, 'data': 0},
            sort_action='custom', sort_mode='multi', sort_by=[],
            active_cell={'row': 0, 'column': 0, 'column_id': '_index_'},
            style_table={'overflowY': 'auto', 'overflowX': 'auto'},
            style_header={'height': '30px'},
            style_cell={
                'minWidth': '70px', 'width': '120px', 'maxWidth': '200px',
            },
        )

        self.components['graph']['global_feature_importance'] = MyGraph(
            figure=go.Figure(), id='global_feature_importance'
        )

        self.components['graph']['feature_selector'] = MyGraph(
            figure=go.Figure(), id='feature_selector'
        )

        # Component for the graph prediction picking
        self.components['graph']['prediction_picking'] = MyGraph(
            figure=go.Figure(), id='prediction_picking'
        )

        self.components['graph']['detail_feature'] = MyGraph(
            figure=go.Figure(), id='detail_feature'
        )

        # Component create to filter the dataset
        self.components['filter']['filter_dataset'] = dbc.Col(
            [dbc.Row(
                html.Div(
                    id='main',
                    children=[
                        html.Div(
                                id='filters',
                                children=[
                                    # Create Add Filter button
                                    dbc.Button(
                                        id='add_dropdown_button',
                                        children='Add Filter',
                                        color='warning',
                                        size='sm',
                                        style={'margin-right': '20px'}
                                    ),
                                    # Create reset Filter button (disabled of no filters applied)
                                    dbc.Button(
                                        id='reset_dropdown_button',
                                        children='Reset all existing filters',
                                        color='warning', disabled=True,
                                        size='sm',
                                        style={'margin-right': '20px'}
                                    ),
                                    # Create explanation button
                                    dbc.Button(
                                        "?",
                                        id="open_filter",
                                        size='sm',
                                        color="warning",
                                        ),
                                    # Create popover on the explanation button
                                    dbc.Popover(
                                        "Click here to know how you can apply filters.",
                                        target="open_filter",
                                        body=True,
                                        trigger="hover",
                                    ),
                                    # Modal associated to the explanation button
                                    dbc.Modal(
                                           [
                                            dbc.ModalHeader(
                                                dbc.ModalTitle("Filters explanation")
                                                ),
                                            dbc.ModalBody([
                                                html.Div(
                                                    dcc.Markdown(
                                                        self.explanations.filter
                                                        )
                                                    )
                                               ]),
                                            dbc.ModalFooter(
                                                dbc.Button(
                                                    "Close",
                                                    id="close_filter",
                                                    color="warning"
                                                    )
                                               ),
                                           ],
                                           id="modal_filter",
                                           centered=True,
                                           size='lg'
                                              ),
                                    # Div which will contains the filters
                                    html.Div(
                                        id='dropdowns_container',
                                        children=[]
                                    )
                                ]
                            )
                        ]
                    )
                ),
                dbc.Row(
                    html.Div([
                        html.Br(),
                        # Create Apply Filter button (Hidden if no filter to apply)
                        dbc.Button(
                                id='apply_filter',
                                children='Apply filters',
                                color='warning',
                                size='sm',
                                style={'display': 'none'}
                                )
                            ],
                        )
                    )
                ], style={'maxheight': '22rem', 'height': '21rem', 'zIndex': 800}
            )

        self.components['filter']['index'] = dbc.Col(dbc.Row(
            [
                dbc.Label("Index", align="center", width=4),
                dbc.Col([
                    dbc.Input(
                        id="index_id", type="text", size="s", placeholder="Id must exist",
                        debounce=True, persistence=True, style={'textAlign': 'right'}
                    )], width={"size": 5},
                    style={'padding': "0px"}
                ),
                dbc.Col([
                    html.Img(id='validation', alt='Validate', title='Validate index',
                             src=self.app.get_asset_url('reload.png'),
                             height='30px', style={'cursor': 'pointer'},
                             )], width={"size": 2},
                        style={'padding': "0px"}, align="center"
                        )
            ])
        )

        self.components['filter']['threshold'] = dbc.Col(
            [
                dbc.Label("Threshold", html_for="slider", id='threshold_label'),
                dcc.Slider(
                    min=0, max=self.max_threshold, value=0, step=0.1,
                    marks={f'{round(self.max_threshold * mark / 4)}': f'{round(self.max_threshold * mark / 4)}'
                           for mark in range(5)},
                    id="threshold_id",
                )
            ],
            className='filter_dashed'
        )

        self.components['filter']['max_contrib'] = dbc.Col(
            [
                dbc.Label(
                    "Features to display: ", id='max_contrib_label'),
                dcc.Slider(
                    min=1, max=min(self.settings['features'], len(self.dataframe.columns) - 2),
                    step=1, value=min(self.settings['features'], len(self.dataframe.columns) - 2),
                    id="max_contrib_id",
                )
            ],
            className='filter_dashed'
        )

        self.components['filter']['positive_contrib'] = dbc.Col(
            [dbc.Row(
                [
                    dbc.Label("Contributions to display:", style={'font-size': '95%'}),
                ]
            ),
                dbc.Row(
                    [
                        dbc.Col(
                            dbc.Checklist(
                                options=[{"label": "Positive", "value": 1}], value=[1], inline=True,
                                id="check_id_positive",
                                # define the font-size style
                                style={'font-size': '82%'}
                            ), style={'display': 'inline-block'}
                        ),
                        dbc.Col(
                            dbc.Checklist(
                                options=[{"label": "Negative", "value": 1}], value=[1], inline=True,
                                id="check_id_negative",
                                # define the font-size style
                                style={'font-size': '82%'}
                                ), style={'display': 'inline-block'}, align="center"
                            )
                    ], className="g-0", justify="center"
                )
            ],
            className='filter_dashed'
        )

        self.components['filter']['masked_contrib'] = dbc.Col(
            [
                dbc.Label(
                    "Feature(s) to mask:"),
                dcc.Dropdown(options=[{'label': key, 'value': value} for key, value in sorted(
                    self.explainer.inv_features_dict.items(), key=lambda item: item[0])],
                    value='', multi=True, searchable=True,
                    id="masked_contrib_id"
                ),
            ],
            className='filter_dashed'
        )

    def make_skeleton(self):
        """
        Describe the app skeleton (bootstrap grid) and initialize components containers
        """
        self.skeleton['navbar'] = dbc.Container(
            [
                dbc.Row([
                        dbc.Col(
                            html.A(
                                dbc.Row([
                                    dbc.Col([
                                        html.Img(src=self.logo, height="40px")], className='col-1'),
                                    dbc.Col([
                                        html.H4("Shapash Monitor", id="shapash_title")]),
                                    ],
                                    align="center", style={'color': self.title_menu_color}
                                ),
                                href="https://github.com/MAIF/shapash", target="_blank",
                            ),
                            # Change md=3 to md=2
                            md=2, align="center", width="100%", style={'padding': 'auto'}
                        ),
                        dbc.Col([
                            html.A(
                                dbc.Row([
                                        html.H3(truncate_str(self.explainer.title_story, maxlen=40),
                                                id="shapash_title_story",
                                                style={'text-align': 'center'})]
                                        ),
                                href="https://github.com/MAIF/shapash", target="_blank",
                            )],
                            # Change md=3 to md=4
                            md=4, align="center", width="100%", style={'padding': 'auto'}
                        ),
                        dbc.Col([
                            self.components['menu']
                            ], align="end", md=6, width='100%'
                        )
                        ],
                        style={'padding': "5px 15px",
                               "verticalAlign": "middle",
                               "width": "auto",
                               "justify": "end"}
                        )
            ],
            fluid=True, style={'height': '100%', 'backgroundColor': self.bkg_color
                               }
        )

        self.skeleton['body'] = dbc.Container(
            [
                dbc.Row(
                    [
                        dbc.Col(
                            [
                                dbc.Card([
                                        html.Div(
                                         # To drow the global_feature_importance graph
                                         self.draw_component('graph', 'global_feature_importance'),
                                         id="card_global_feature_importance",
                                         # Position must be absolute to add the explanation button
                                         style={"position": 'absolute'}
                                         ),
                                        html.Div([
                                             # Create explanation button on feature importance graph
                                             dbc.Button(
                                                 "?",
                                                 id="open_feature_importance",
                                                 size='sm',
                                                 color="warning"
                                                 ),
                                             # Create popover for this button
                                             dbc.Popover(
                                                  "Click here to have more \
                                                  information on Feature Importance graph.",
                                                  target="open_feature_importance",
                                                  body=True,
                                                  trigger="hover",
                                                     ),
                                             # Create modal associated to this button
                                             dbc.Modal([
                                                        # Modal title
                                                        dbc.ModalHeader(
                                                            dbc.ModalTitle("Feature importance")
                                                            ),
                                                        dbc.ModalBody([
                                                            html.Div(
                                                                # Add explanation
                                                                dcc.Markdown(
                                                                    self.explanations.feature_importance
                                                                    )
                                                                    ),
                                                            # Here to add link in the modal
                                                            html.A('Click here for more details',
                                                                   href="https://github.com/MAIF/shapash/blob/master/tutorial/plot/tuto-plot03-features-importance.ipynb",
                                                                   # open new brother tab
                                                                   target="_blank",
                                                                   style={'color': self.color[0]})
                                                        ]),
                                                        # button to close the modal
                                                        dbc.ModalFooter(
                                                            dbc.Button(
                                                                "Close",
                                                                id="close_feature_importance",
                                                                color="warning"
                                                            )
                                                        ),
                                                    ],
                                                    id="modal_feature_importance",
                                                    centered=True,
                                                    size='lg'
                                                )
                                                ],
                                                # position must be relative
                                                style={'position': 'relative', 'left': '96%'})
                                    ])
                            ],
                            md=5,
                            style={'padding': '10px 10px 0px 10px'},
                        ),
                        dbc.Col(
                            [
                                # Tabs that contain 3 children tab (Dataset,
                                # Dataset Filters and True Value Vs Pedicted Values)
                                dbc.Tabs([
                                    # Tab which contains the datatable component
                                    dbc.Tab(
                                        # draw datatable component
                                        self.draw_component('table', 'dataset'),
                                        # Tab name
                                        label='Dataset',
                                        className="card",
                                        id='card_dataset',
                                        # Style of the tab
                                        style={'cursor': 'pointer'},
                                        label_style={'color': "black", 'height': '30px',
                                                     'padding': '0px 5px'},
                                        # Style when the tab is activated
                                        active_tab_class_name="fw-bold fst-italic",
                                        active_label_style={'border-top': '3px solid',
                                                            'border-top-color': self.color[0]
                                                            }
                                        ),
                                    # Tab which contains components to filter the dataset
                                    dbc.Tab(
                                        dbc.Card(
                                            dbc.CardBody(
                                                html.Div(
                                                    # draw the component
                                                    self.draw_filter_table(),
                                                    id='card_filter_dataset',
                                                    # To add scroll in overflow y
                                                    style={'overflow-y': "scroll",
                                                           'overflow-x': 'hidden'})
                                            ), style={'height': '24.1rem'},
                                        ),
                                        # Tab name
                                        label='Dataset Filters',
                                        # Style of the tab
                                        label_style={'color': "black", 'height': '30px',
                                                     'padding': '0px 5px'},
                                        tab_style={'border-left': '2px solid #ddd',
                                                   'border-right': '2px solid #ddd'},
                                        # Style when the tab is activated
                                        active_tab_class_name="fw-bold fst-italic",
                                        active_label_style={'border-top': '3px solid',
                                                            'border-top-color': self.color[0]
                                                            }
                                         ),
                                    # Tab which contains prediction picking graph
                                    # and its explanation button
                                    dbc.Tab(
                                        dbc.Card([
                                            html.Div(
                                                # draw prediction picking graph
                                                self.draw_component('graph', 'prediction_picking'),
                                                id="card_prediction_picking",
                                                # Position must be absolute to add
                                                # the explanation button
                                                style={"position": 'absolute'}
                                            ),
                                            html.Div([
                                                 # Create explanation button
                                                 dbc.Button(
                                                     "?",
                                                     id="open_prediction_picking",
                                                     size='sm',
                                                     color="warning"
                                                     ),
                                                 # Create popover for this button
                                                 dbc.Popover(
                                                         "Click here to have more \
                                                         information on Prediction Picking graph.",
                                                         target="open_prediction_picking",
                                                         body=True,
                                                         trigger="hover",
                                                     ),
                                                 # Create modal associated to this button
                                                 dbc.Modal([
                                                             # Modal title
                                                             dbc.ModalHeader(
                                                                dbc.ModalTitle("True values Vs Predicted values")
                                                              ),
                                                             dbc.ModalBody([
                                                                    html.Div(
                                                                        # explanation
                                                                        dcc.Markdown(self.explanations.prediction_picking)
                                                                        ),
                                                                    # Here to add link in the modal
                                                                    html.Div(html.Img(src="https://github.com/MAIF/shapash/blob/master/docs/_static/shapash_select_subset.gif?raw=true")),
                                                                    # Here to add a link in the modal
                                                                    html.A('Click here for more details',
                                                                           href="https://github.com/MAIF/shapash/blob/master/tutorial/plot/tuto-plot06-prediction_plot.ipynb",
                                                                           # open new brother tab
                                                                           target="_blank",
                                                                           style={'color': self.color[0]})
                                                                    ]),
                                                             dbc.ModalFooter(
                                                                 # button to close the modal
                                                                 dbc.Button(
                                                                     "Close",
                                                                     id="close_prediction_picking",
                                                                     color="warning"
                                                                 )
                                                              ),
                                                        ],
                                                        id="modal_prediction_picking",
                                                        centered=True,
                                                        size='lg'
                                                    )
                                                 # Position must be relative
                                                ],  style={'position': 'relative', 'left': '97%'})
                                         ]),
                                        # Tab name
                                        label='True Values Vs Predicted Values',
                                        # Style of the tab
                                        label_style={'color': "black", 'height': '30px',
                                                     'padding': '0px 5px'},
                                        # Style when the tab is activated
                                        active_tab_class_name="fw-bold fst-italic",
                                        active_label_style={'border-top': '3px solid',
                                                            'border-top-color': self.color[0]
                                                            }
                                    )
                                ], id="tabs"
                                )
                            ],
                            md=7,
                            style={'padding': '10px 10px'},
                        ),
                    ], style={'padding': '10px 10px 0px 10px',
                              'height': '100%'}, align="center"
                ),
                dbc.Row(
                    [
                        dbc.Col(
                            [
                                # Card which contains feature selector graph
                                # and explanation button
                                dbc.Card([
                                        html.Div(
                                            # Draw feature selector graph
                                            self.draw_component('graph', 'feature_selector'),
                                            id='card_feature_selector',
                                            # Position must be absolute to
                                            # add explanation button
                                            style={"position": 'absolute'}
                                             ),
                                         html.Div([
                                             # Create explanation button
                                             dbc.Button(
                                                 "?",
                                                 id="open_feature_selector",
                                                 size='sm',
                                                 color="warning"
                                                 ),
                                             # popover of this button
                                             dbc.Popover(
                                                         "Click here to have more \
                                                         information on Feature Selector graph.",
                                                         target="open_feature_selector",
                                                         body=True,
                                                         trigger="hover",
                                                     ),
                                             # Modal of this button
                                             dbc.Modal([
                                                        dbc.ModalHeader(
                                                            dbc.ModalTitle("Feature selector")
                                                            ),
                                                        dbc.ModalBody([
                                                            html.Div(
                                                                # explanations
                                                                dcc.Markdown(self.explanations.feature_selector)
                                                                ),
                                                            # Here to add link
                                                            html.A('Click here for more details',
                                                                   href="https://github.com/MAIF/shapash/blob/master/tutorial/plot/tuto-plot02-contribution_plot.ipynb",
                                                                   # open new brother tab
                                                                   target="_blank",
                                                                   style={'color': self.color[0]})
                                                                ]),
                                                        dbc.ModalFooter(
                                                            # Button to close modal
                                                            dbc.Button(
                                                                "Close",
                                                                id="close_feature_selector",
                                                                color="warning"
                                                            )
                                                        ),
                                                    ],
                                                    id="modal_feature_selector",
                                                    centered=True,
                                                    size='lg'
                                                )
                                                ],
                                                 # position must be relative
                                                 style={'position': 'relative',
                                                        'left': '96%'})
                                          ])
                            ],
                            md=5,
                            align="center",
                            style={'padding': '0px 10px'},
                        ),
                        dbc.Col(
                            [
                                dbc.Row(
                                    [
                                        dbc.Col(
                                            [
                                             # Card that contains detail feature graph
                                             # and explanation button
                                             dbc.Card([
                                                 html.Div(
                                                     # draw detail_feature graph
                                                     self.draw_component('graph', 'detail_feature'),
                                                     id='card_detail_feature',
                                                     style={"position": 'absolute'}
                                                 ),
                                                 html.Div([
                                                     # Create explanation button
                                                     dbc.Button(
                                                         "?",
                                                         id="open_detail_feature",
                                                         size='sm',
                                                         color="warning"
                                                         ),
                                                     # Popover of this button
                                                     dbc.Popover(
                                                         "Click here to have more \
                                                         information on Detail Feature graph.",
                                                         target="open_detail_feature",
                                                         body=True,
                                                         trigger="hover",
                                                     ),
                                                     # Modal of this button
                                                     dbc.Modal(
                                                            [
                                                             dbc.ModalHeader(
                                                                 dbc.ModalTitle("Detail feature")
                                                                 ),
                                                             dbc.ModalBody([
                                                                 html.Div(
                                                                     # explanations
                                                                     dcc.Markdown(self.explanations.detail_feature)
                                                                     ),
                                                                 # Here to add link on the modal
                                                                 html.A('Click here for more details',
                                                                        href="https://github.com/MAIF/shapash/blob/master/tutorial/plot/tuto-plot01-local_plot-and-to_pandas.ipynb",
                                                                        # open new brother tab
                                                                        target="_blank",
                                                                        style={'color': self.color[0]})
                                                                ]),
                                                             dbc.ModalFooter(
                                                                 # Button to close the modal
                                                                 dbc.Button(
                                                                     "Close",
                                                                     id="close_detail_feature",
                                                                     color="warning"
                                                                     )
                                                                ),
                                                            ],
                                                            id="modal_detail_feature",
                                                            centered=True,
                                                            size='lg'
                                                        )

                                                        ],
                                                          # Position must be relative
                                                          style={
                                                            'position': 'relative',
                                                            'left': '96%'
                                                            })
                                                ])
                                            ],
                                            md=8,
                                            align="center",
                                        ),
                                        dbc.Col(
                                            [
                                                html.Div(
                                                    self.draw_filter(),
                                                    className="card_filter",
                                                    id='card_filter',
                                                ),
                                            ],
                                            md=4,
                                            align="center",
                                        ),
                                    ],
                                ),
                            ],
                            md=7,
                            align="center",
                            style={'padding': '0px 10px'},
                        ),
                    ],
                    style={'padding': '10px 5px 10px 10px'},
                ),
            ],
            className="mt-12",
            fluid=True,
            # To drop the x scroll-bar
            style={'overflow-x': 'hidden'}
        )

    def adjust_menu(self):
        """
        Override menu from explainer object depending on
        classification or regression case.
        """
        on_style = {'backgroundColor': self.color[0],
                    'color': self.bkg_color,
                    'margin-right': '0.5rem'}
        off_style = {'backgroundColor': self.color[1],
                     'color': self.bkg_color,
                     'margin-right': '0.5rem'}
        if self.explainer._case == 'classification':
            self.components['menu']['select_label'].options = \
                [
                    {'label': f'{self.explainer.label_dict[label] if self.explainer.label_dict else label}',
                     'value': label}
                    for label in self.explainer._classes
            ]
            self.components['menu']['classification_badge'].style = on_style
            self.components['menu']['regression_badge'].style = off_style
            self.components['menu']['select_label'].value = self.label

        elif self.explainer._case == 'regression':
            self.components['menu']['classification_badge'].style = off_style
            self.components['menu']['regression_badge'].style = on_style
            self.components['menu']['select_collapse'].is_open = False

        else:
            raise ValueError(f'No rule defined for explainer case : {self.explainer._case}')

    def draw_component(self,
                       component_type,
                       component_id,
                       title=None):
        """
        Method which return a component from a type and id.
        It's the method to insert component inside component container.
        Parameters
        ----------
        component_type : string
            Type of the component. Can be table, graph, ...
        component_id : string
            Id of the component. It must be unique.
        title : string, optional
            by default None
        Returns
        -------
        list
            list of components
            (combining for example Graph + embed button to get fullscreen
             details)
        """
        component = [html.H4(title)] if title else []
        component.append(self.components[component_type][component_id])
        component.append(
            html.A(
                html.I("fullscreen",
                       className="material-icons tiny",
                       style={'marginTop': '8px', 'marginLeft': '1px'}
                       ),
                id=f"ember_{component_id}",
                className="dock-expand",
                **{'data-component-type': component_type},
                # Get components'id
                **{'data-component-id': component_id}
            )
        )
        return component
    
    def draw_filter_table(self):
        """
        Method which returns the filter dataset components block.
        Returns
        -------
            component
        """
        return self.components['filter']['filter_dataset']
    
    def draw_filter(self):
        """
        Method which returns filter components block for local
        contributions plot.
        Returns
        -------
        list
            list of components
        """
        filter = [
            dbc.Container(
                [
                    dbc.Row([self.components['filter']['index']],
                            align="center", style={"height": "4rem"}
                            ),
                    dbc.Row([self.components['filter']['threshold']],
                            align="center", style={"height": "5rem"}
                            ),
                    dbc.Row([self.components['filter']['max_contrib']],
                            align="center", style={"height": "5rem"}
                            ),
                    dbc.Row([self.components['filter']['positive_contrib']],
                            align="center", style={"height": "4rem"}
                            ),
                    dbc.Row([self.components['filter']['masked_contrib']],
                            align="center"),
                ],
            ),
        ]
        return filter

    def select_point(self,
                     graph,
                     click_data):
        """
        Method which set the selected point in graph component
        corresponding to click_data.
        """
        if click_data:
            curve_id = click_data['points'][0]['curveNumber']
            point_id = click_data['points'][0]['pointIndex']
            for curve in range(
                    len(self.components['graph'][graph].figure['data'])):
                self.components['graph'][graph].figure['data'][curve].selectedpoints = \
                    [point_id] if curve == curve_id else []

    def callback_fullscreen_buttons(self):
        """
        Initialize callbacks for each fullscreen button
        the callback alter style of the component (height, ...)
        Returns
        -------
        dict
            Style of the component
        """
        app = self.app
        components_to_init = dict([(graph, 'graph') for graph in self.components['graph'].keys()])
        components_to_init['dataset'] = 'table'
        for component_id, component_type in components_to_init.items():
            component_property = 'style' if component_type == "graph" else "style_table"

            @app.callback(
                [
                    Output(f'card_{component_id}', 'style'),
                    Output(f'{component_id}', component_property),
                ],
                [
                    Input(f'ember_{component_id}', 'n_clicks'),
                    Input(f'ember_{component_id}', 'data-component-type'),
                    Input(f'ember_{component_id}', 'data-component-id')
                ]
            )
            def ember(click,
                      data_component_type,
                      data_component_id):
                """
                Function used to set style of cards and components.
                Prediction picking graph style is different than the other
                graph because it is placed in a tab.
                ---------------------------------------------------------------
                click: click on zoom button
                data_component_type: component type
                data_component_id: component id
                --------------------------------------------------------------
                return style of cards and style of components
                """
                click = 2 if click is None else click
                toggle_on = True if click % 2 == 0 else False
                if toggle_on:
                    # Style for graph
                    style_component = {
                        'height': '21.6rem'
                    }
                    this_style_card = {
                        'height': '22rem', 'zIndex': 900,
                    }
                    # Style for prediction picking graph
                    if data_component_id == 'prediction_picking':
                        style_component = {
                            'height': '20.6rem',
                        }
                        this_style_card = {
                            'height': '20.8rem', 'zIndex': 901,
                        }
                    # Style for the Dataset
                    if data_component_type == 'table':
                        style_component = {
                            'maxHeight': '23rem',
                        }
                        this_style_card = {
                            'height': '24.1rem', 'zIndex': 900,
                        }
                    return this_style_card, style_component

                else:
                    # Style when zoom button is clicked
                    this_style_card = {
                        'height': '70vh',
                        'width': 'auto',
                        'zIndex': 998,
                        'position': 'fixed', 'top': '55px',
                        'bottom': 0, 'left': 0, 'right': 0,
                    }
                    style_component = {
                        'height': '89vh', 'maxHeight': '89vh',
                    }
                    return this_style_card, style_component

    def init_callback_settings(self):
        app = self.app
        self.components['settings']['input_rows']['rows'].value = self.settings['rows']
        self.components['settings']['input_points']['points'].value = self.settings['points']
        self.components['settings']['input_features']['features'].value = self.settings['features']
        self.components['settings']['input_violin']['violin'].value = self.settings['violin']

        for id in self.settings.keys():
            @app.callback(
                [Output(f'{id}', 'valid'),
                 Output(f'{id}', 'invalid')],
                [Input(f'{id}', "value")]
            )
            def update_valid(value):
                """
                actualise valid and invalid icon in input component
                Parameters
                ----------
                value : int
                    value of input component
                Returns
                -------
                    tuple of boolean
                """
                patt = re.compile('^[0-9]*[1-9][0-9]*$')
                if patt.match(str(value)):
                    return True, False
                else:
                    return False, True

        @app.callback(
            Output("modal", "is_open"),
            [
                Input("settings", "n_clicks"),
                Input("apply", "n_clicks")],
            [
                State('rows', 'valid'),
                State('points', 'valid'),
                State('features', 'valid'),
                State('violin', 'valid'),
            ],
        )
        def toggle_modal(n1,
                         n2,
                         rows,
                         points,
                         features,
                         violin):
            """
            open modal /close modal (only if all input are valid)
            """
            ctx = dash.callback_context
            if ctx.triggered[0]['prop_id'] == 'settings.n_clicks':
                if n1 is not None:
                    return True
            else:
                if n2 is not None:
                    if all([rows, points, features, violin]):
                        return False
                    else:
                        return True
            return False

    def callback_generator(self):
        app = self.app

        @app.callback(
            [
                Output('dataset', 'data'),
                Output('dataset', 'tooltip_data'),
                Output('dataset', 'columns'),
                Output('dataset', 'active_cell'),
            ],
            [
                Input('prediction_picking', 'selectedData'),
                Input('modal', 'is_open'),
                Input('apply_filter', 'n_clicks'),
                Input('reset_dropdown_button', 'n_clicks'),
                Input({'type': 'del_dropdown_button', 'index': ALL}, 'n_clicks')
            ],
            [
                State('rows', 'value'),
                State('name', 'value'),
                State({'type': 'var_dropdown', 'index': ALL}, 'value'),
                State({'type': 'var_dropdown', 'index': ALL}, 'id'),
                State({'type': 'dynamic-str', 'index': ALL}, 'value'),
                State({'type': 'dynamic-str', 'index': ALL}, 'id'),
                State({'type': 'dynamic-bool', 'index': ALL}, 'value'),
                State({'type': 'dynamic-bool', 'index': ALL}, 'id'),
                State({'type': 'dynamic-date', 'index': ALL}, 'start_date'),
                State({'type': 'dynamic-date', 'index': ALL}, 'end_date'),
                State({'type': 'dynamic-date', 'index': ALL}, 'id'),
                State({'type': 'lower', 'index': ALL}, 'value'),
                State({'type': 'lower', 'index': ALL}, 'id'),
                State({'type': 'upper', 'index': ALL}, 'value'),
                State({'type': 'upper', 'index': ALL}, 'id'),
                State('dropdowns_container', 'children')
            ]
        )
        def update_datatable(selected_data,
                             is_open,
                             nclicks_apply,
                             nclicks_reset,
                             nclicks_del,
                             rows,
                             name,
                             val_feature,
                             id_feature,
                             val_str_modality,
                             id_str_modality,
                             val_bool_modality,
                             id_bool_modality,
                             start_date,
                             end_date,
                             id_date,
                             val_lower_modality,
                             id_lower_modality,
                             val_upper_modality,
                             id_upper_modality,
                             children):
            """
            This function is used to update the datatable according to sorting,
            filtering and settings modifications.
            ------------------------------------------------------------------
            selected_data: selected data in prediction picking graph
            is_open: modal
            nclicks_apply: click on Apply Filter button
            nclicks_reset: click on Reset All Filter button
            nclicks_del: click on delete button
            rows: number of rows for subset
            name: name for features name
            val_feature: feature selected to filter
            id_feature: id of feature selected to filter
            val_str_modality: string modalities selected
            id_str_modality: id of string modalities selected
            val_bool_modality: boolean modalities selected
            id_bool_modality: id of boolean modalities selected
            start_date: start dates selected
            end_date: end dates selected
            id_date: id of dates selected
            val_lower_modality: lower values of numeric filter
            id_lower_modality: id of lower modalities of numeric filter
            val_upper_modality: upper values of numeric filter
            id_upper_modality: id of upper values of numeric filter
            children: children of dropdown container
            ------------------------------------------------------------------
            return
            data: available dataset
            tooltip_data: tooltip of the dataset
            columns: columns of the dataset
            active_cell: activated cell
            """
            ctx = dash.callback_context
            active_cell = no_update
            df = self.round_dataframe
            columns = self.components['table']['dataset'].columns
            if ctx.triggered[0]['prop_id'] == 'modal.is_open':
                if is_open:
                    raise PreventUpdate
                else:
                    self.settings['rows'] = rows
                    self.init_data()
                    active_cell = {'row': 0, 'column': 0, 'column_id': '_index_'}
                    self.settings_ini['rows'] = self.settings['rows']
                    if name == [1]:
                        columns = [
                            {"name": '_index_', "id": '_index_'},
                            {"name": '_predict_', "id": '_predict_'}] + \
                            [{"name": self.explainer.features_dict[i], "id": i} for i in self.explainer.x_init]
                    df = self.round_dataframe
            elif ((ctx.triggered[0]['prop_id'] == 'prediction_picking.selectedData') and
                  (selected_data is not None) and (len(selected_data) > 1)):
                row_ids = []
                # If some data have been selected in prediction picking graph
                if selected_data is not None and len(selected_data) > 1:
                    for p in selected_data['points']:
                        row_ids.append(p['customdata'])
                    df = self.round_dataframe.loc[row_ids]
                else:
                    df = self.round_dataframe
            # If click on reset button
            elif ctx.triggered[0]['prop_id'] == 'reset_dropdown_button.n_clicks':
                df = self.round_dataframe
            # If click on Apply filter
            elif ((ctx.triggered[0]['prop_id'] == 'apply_filter.n_clicks') | (
                    (ctx.triggered[0]['prop_id'] == 'prediction_picking.selectedData') and
                  ((selected_data is None))) | (
                    (ctx.triggered[0]['prop_id'] == 'prediction_picking.selectedData') and
                  (selected_data is not None and len(selected_data) == 1 and selected_data['points'][0]['curveNumber'] > 0)
                  )):
                # get list of ID
                feature_id = [id_feature[i]['index'] for i in range(len(id_feature))]
                str_id = [id_str_modality[i]['index'] for i in range(len(id_str_modality))]
                bool_id = [id_bool_modality[i]['index'] for i in range(len(id_bool_modality))]
                lower_id = [id_lower_modality[i]['index'] for i in range(len(id_lower_modality))]
                date_id = [id_date[i]['index'] for i in range(len(id_date))]
                df = self.round_dataframe
                # If there is some filters
                if len(feature_id) > 0:
                    for i in range(len(feature_id)):
                        # String filter
                        if feature_id[i] in str_id:
                            position = np.where(np.array(str_id) == feature_id[i])[0][0]
                            if ((position is not None) & (val_str_modality[position] is not None)):
                                df = df[df[val_feature[i]].isin(val_str_modality[position])]
                            else:
                                df = df
                        # Boolean filter
                        elif feature_id[i] in bool_id:
                            position = np.where(np.array(bool_id) == feature_id[i])[0][0]
                            if ((position is not None) & (val_bool_modality[position] is not None)):
                                df = df[df[val_feature[i]] == val_bool_modality[position]]
                            else:
                                df = df
                        # Date filter
                        elif feature_id[i] in date_id:
                            position = np.where(np.array(date_id) == feature_id[i])[0][0]
                            if((position is not None) &
                               (start_date[position] < end_date[position])):
                                df = df[((df[val_feature[i]] >= start_date[position]) &
                                         (df[val_feature[i]] <= end_date[position]))]
                            else:
                                df = df
                        # Numeric filter
                        elif feature_id[i] in lower_id:
                            position = np.where(np.array(lower_id) == feature_id[i])[0][0]
                            if((position is not None) & (val_lower_modality[position] is not None) &
                               (val_upper_modality[position] is not None)):
                                if (val_lower_modality[position] < val_upper_modality[position]):
                                    df = df[(df[val_feature[i]] >= val_lower_modality[position]) &
                                            (df[val_feature[i]] <= val_upper_modality[position])]
                                else:
                                    df = df
                            else:
                                df = df
                        else:
                            df = df
                else:
                    df = df
                if len(df) == 0:
                    raise ValueError(
                        "Your dataframe is empty. It must have at list one row"
                         )
            elif None not in nclicks_del:
                df = self.round_dataframe
            else:
                raise dash.exceptions.PreventUpdate
            self.components['table']['dataset'].data = df.to_dict('records')
            self.components['table']['dataset'].tooltip_data = [
                {
                    column: {'value': str(value), 'type': 'text'}
                    for column, value in row.items()
                } for row in df.to_dict('rows')
            ]
            return (
                self.components['table']['dataset'].data,
                self.components['table']['dataset'].tooltip_data,
                columns,
                active_cell,
            )

        @app.callback(
            [
                Output('global_feature_importance', 'figure'),
                Output('global_feature_importance', 'clickData')
            ],
            [
                Input('select_label', 'value'),
                Input('dataset', 'data'),
                Input('prediction_picking', 'selectedData'),
                Input('apply_filter', 'n_clicks'),
                Input('reset_dropdown_button', 'n_clicks'),
                Input({'type': 'del_dropdown_button', 'index': ALL}, 'n_clicks'),
                Input('modal', 'is_open'),
                Input('card_global_feature_importance', 'n_clicks'),
                Input('bool_groups', 'on'),
                Input('ember_global_feature_importance', 'n_clicks')
            ],
            [
                State('global_feature_importance', 'clickData'),
                State('features', 'value')
            ]
        )
        def update_feature_importance(label,
                                      data,
                                      selected_data,
                                      apply_filters,
                                      reset_filter,
                                      nclicks_del,
                                      is_open,
                                      n_clicks,
                                      bool_group,
                                      click_zoom,
                                      clickData,
                                      features):
            """
            update feature importance plot according label, click on graph,
            filters applied and subset selected in prediction picking graph.
            ------------------------------------------------------------
            label: label of data
            data: dataset
            selected_data : data selected on prediction picking graph
            apply_filters: click on apply filter button
            reset_filter: click on reset filter button
            nclicks_del: click on del button
            is_open: modal
            n_clicks: click on features importance card
            bool_group: display groups
            click_zoom: click on zoom button
            clickData: click on features importance graph
            features: features value
            -------------------------------------------------------------
            return
            figure of Features Importance graph
            click on Features Importance graph
            """
            ctx = dash.callback_context
            # Zoom is False by Default. It becomes True if we click on it
            click = 2 if click_zoom is None else click_zoom
            if click % 2 == 0:
                zoom_active = False
            else:
                zoom_active = True
            selection = None
            selected_feature = self.explainer.inv_features_dict.get(
                clickData['points'][0]['label'].replace('<b>', '').replace('</b>', '')
            ) if clickData else None
            if ctx.triggered[0]['prop_id'] == 'modal.is_open':
                if is_open:
                    raise PreventUpdate
                else:
                    self.settings['features'] = features
                    self.settings_ini['features'] = self.settings['features']
            elif ctx.triggered[0]['prop_id'] == 'select_label.value':
                self.label = label
                selection = None
            elif ctx.triggered[0]['prop_id'] == 'dataset.data':
                self.list_index = [d['_index_'] for d in data]
            elif ctx.triggered[0]['prop_id'] == 'bool_groups.on':
                clickData = None  # We reset the graph and clicks if we toggle the button
            # If we have selected data on prediction picking graph
            elif ((ctx.triggered[0]['prop_id'] == 'prediction_picking.selectedData') and
                  (selected_data is not None) and (len(selected_data) > 1)):
                row_ids = []
                if selected_data is not None and len(selected_data) > 1:
                    for p in selected_data['points']:
                        row_ids.append(p['customdata'])
                    selection = row_ids
                else:
                    selection = None
                #when group
                if self.explainer.features_groups and bool_group:
                    list_sub_features = [f for group_features in self.explainer.features_groups.values()
                                      for f in group_features]
                    if selected_feature not in list_sub_features:
                        selected_feature = None
            # If click on a single point on prediction picking, do nothing
            elif ((ctx.triggered[0]['prop_id'] == 'prediction_picking.selectedData') and
                  (selected_data is not None) and (len(selected_data) == 1)):
                # If there is some filters applied
                if (len([d['_index_'] for d in data]) != len(self.list_index)):
                    selection = [d['_index_'] for d in data]
                else:
                    selection = None
            # If we have dubble click on prediction picking to remove the selected subset
            elif ((ctx.triggered[0]['prop_id'] == 'prediction_picking.selectedData') and
                  (selected_data is None)):
                # If there is some filters applied
                if (len([d['_index_'] for d in data]) != len(self.list_index)):
                    selection = [d['_index_'] for d in data]
                else:
                    selection = None
                #when group
                if self.explainer.features_groups and bool_group:
                    list_sub_features = [f for group_features in self.explainer.features_groups.values()
                                      for f in group_features]
                    if selected_feature not in list_sub_features:
                        selected_feature = None
            # If we click on reset filter button
            elif ctx.triggered[0]['prop_id'] == 'reset_dropdown_button.n_clicks':
                selection = None
                #when group
                if self.explainer.features_groups and bool_group:
                    list_sub_features = [f for group_features in self.explainer.features_groups.values()
                                      for f in group_features]
                    if selected_feature not in list_sub_features:
                        selected_feature = None
            # If we click on Apply button
            elif ctx.triggered[0]['prop_id'] == 'apply_filter.n_clicks':
                selection = [d['_index_'] for d in data]
                #when group
                if self.explainer.features_groups and bool_group:
                    list_sub_features = [f for group_features in self.explainer.features_groups.values()
                                      for f in group_features]
                    if selected_feature not in list_sub_features:
                        selected_feature = None
            # If we click on the last del button
            elif (('del_dropdown_button' in ctx.triggered[0]['prop_id']) &
                  (None not in nclicks_del)):
                selection = None
                #when group
                if self.explainer.features_groups and bool_group:
                    list_sub_features = [f for group_features in self.explainer.features_groups.values()
                                      for f in group_features]
                    if selected_feature not in list_sub_features:
                        selected_feature = None
            elif (ctx.triggered[0]['prop_id'] == 'card_global_feature_importance.n_clicks'
                  and self.explainer.features_groups and bool_group):
                row_ids = []
                if selected_data is not None and len(selected_data) > 1:
                    # we plot prediction picking subset
                    for p in selected_data['points']:
                        row_ids.append(p['customdata'])
                    selection = row_ids
                elif (len([d['_index_'] for d in data]) != len(self.list_index)):
                    selection = [d['_index_'] for d in data]
                else:
                    selection = None
                # When we click twice on the same bar this will reset the graph
                if self.last_click_data == clickData:
                    selected_feature = None
                list_sub_features = [f for group_features in self.explainer.features_groups.values()
                                      for f in group_features]
                if selected_feature in list_sub_features:
                    self.last_click_data = clickData
                    raise PreventUpdate
                else:
                    pass
            else:
                # Zoom management to generate graph which have global axis
                if len(self.components['graph']['global_feature_importance'].figure['data']) == 1:
                    selection = None
                else:
                    row_ids = []
                    if selected_data is not None and len(selected_data) > 1:
                        # we plot prediction picking subset
                        for p in selected_data['points']:
                            row_ids.append(p['customdata'])
                        selection = row_ids
                    else:
                        # we plot filter subset
                        selection = [d['_index_'] for d in data]
                self.last_click_data = clickData

            group_name = selected_feature if (self.explainer.features_groups is not None
                                              and selected_feature in self.explainer.features_groups.keys()) else None

            self.components['graph']['global_feature_importance'].figure = \
                self.explainer.plot.features_importance(
                    max_features=features,
                    selection=selection,
                    label=self.label,
                    group_name=group_name,
                    display_groups=bool_group,
                    zoom=zoom_active
                )
            # Adjust graph with adding x axis title
            self.components['graph']['global_feature_importance'].adjust_graph(x_ax='Contribution')
            self.components['graph']['global_feature_importance'].figure.layout.clickmode = 'event+select'
            if selected_feature:
                if self.explainer.features_groups is None:
                    self.select_point('global_feature_importance', clickData)
                elif selected_feature not in self.explainer.features_groups.keys():
                    self.select_point('global_feature_importance', clickData)

            # font size can be adapted to screen size
            nb_car = max([len(self.components['graph']['global_feature_importance'].figure.data[0].y[i]) for i in
                          range(len(self.components['graph']['global_feature_importance'].figure.data[0].y))])
            self.components['graph']['global_feature_importance'].figure.update_layout(
                yaxis=dict(tickfont={'size': min(round(500 / nb_car), 12)})
            )

            self.last_click_data = clickData
            return self.components['graph']['global_feature_importance'].figure, clickData

        @app.callback(
            Output(component_id='feature_selector', component_property='figure'),
            [
                Input('global_feature_importance', 'clickData'),
                Input('prediction_picking', 'selectedData'),
                Input('dataset', 'data'),
                Input('apply_filter', 'n_clicks'),
                Input('reset_dropdown_button', 'n_clicks'),
                Input({'type': 'del_dropdown_button', 'index': ALL}, 'n_clicks'),
                Input('select_label', 'value'),
                Input('modal', 'is_open'),
                Input('ember_feature_selector', 'n_clicks')
            ],
            [
                State('points', 'value'),
                State('violin', 'value')
            ]
        )
        def update_feature_selector(feature,
                                    selected_data,
                                    data,
                                    apply_filters,
                                    reset_filter,
                                    nclicks_del,
                                    label,
                                    is_open,
                                    click_zoom,
                                    points,
                                    violin):
            """
            Update feature plot according to label, data,
            selected feature on features importance graph,
            filters and settings modifications
            --------------------------------------------
            feature: click on feature importance graph
            selected_data: Data selected on prediction picking graph
            data: dataset
            apply_filters: click on apply filter button
            reset_filter: click on reset filter button
            nclicks_del: click del button
            label: selected label
            is_open: modal
            click_zoom: click on zoom button
            points: points value in setting
            violin: violin value in setting
            ---------------------------------------------
            return
            figure: feature selector graph
            """
            # Zoom is False by Default. It becomes True if we click on it
            click = 2 if click_zoom is None else click_zoom
            if click % 2 == 0:
                zoom_active = False
            else:
                zoom_active = True  # To check if zoom is activated
            ctx = dash.callback_context
            if ctx.triggered[0]['prop_id'] == 'modal.is_open':
                if is_open:
                    raise PreventUpdate
                else:
                    self.settings['points'] = points
                    self.settings_ini['points'] = self.settings['points']
                    self.settings['violin'] = violin
                    self.settings_ini['violin'] = self.settings['violin']

            elif ctx.triggered[0]['prop_id'] == 'select_label.value':
                self.label = label
            elif ctx.triggered[0]['prop_id'] == 'global_feature_importance.clickData':
                if feature is not None:
                    # Removing bold
                    self.selected_feature = feature['points'][0]['label'].replace('<b>', '').replace('</b>', '')
                    if feature['points'][0]['curveNumber'] == 0 and \
                              len(self.components['graph']['global_feature_importance'].figure['data']) == 2:
                        if selected_data is not None and len(selected_data) > 1:
                            row_ids = []
                            for p in selected_data['points']:
                                row_ids.append(p['customdata'])
                            self.subset = row_ids
                        else:
                            self.subset = [d['_index_'] for d in data]
                    else:
                        self.subset = self.list_index
            # If we have selected data on prediction picking graph
            elif ((ctx.triggered[0]['prop_id'] == 'prediction_picking.selectedData') and
                  (selected_data is not None)):
                row_ids = []
                if selected_data is not None and len(selected_data) > 1:
                    for p in selected_data['points']:
                        row_ids.append(p['customdata'])
                    self.subset = row_ids
            # if we have click on reset button
            elif ctx.triggered[0]['prop_id'] == 'reset_dropdown_button.n_clicks':
                self.subset = None
            # If we have clik on Apply filter button
            elif ctx.triggered[0]['prop_id'] == 'apply_filter.n_clicks':
                self.subset = [d['_index_'] for d in data]
            # If we have click on the last del button
            elif (('del_dropdown_button' in ctx.triggered[0]['prop_id']) &
                  (None not in nclicks_del)):
                self.subset = None
            else:
                # Zoom management to generate graph which have global axis
                if len(self.components['graph']['global_feature_importance'].figure['data']) == 1:
                    self.subset = self.list_index
                elif (len(self.components['graph']['global_feature_importance'].figure['data']) == 2):
                    if feature is not None:
                        if feature['points'][0]['curveNumber'] == 0:
                            if selected_data is not None and len(selected_data) > 1:
                                row_ids = []
                                for p in selected_data['points']:
                                    row_ids.append(p['customdata'])
                                self.subset = row_ids
                            else:
                                self.subset = [d['_index_'] for d in data]
                        else:
                            self.subset = self.list_index
                    else:
                        self.subset = [d['_index_'] for d in data]
                else:
                    row_ids = []
                    if selected_data is not None and len(selected_data) > 1:
                        # we plot prediction picking subset
                        for p in selected_data['points']:
                            row_ids.append(p['customdata'])
                        self.subset = row_ids
                    else:
                        # we plot filter subset
                        self.subset = [d['_index_'] for d in data]

            self.components['graph']['feature_selector'].figure = \
                self.explainer.plot.contribution_plot(
                    col=self.selected_feature,
                    selection=self.subset,
                    label=self.label,
                    violin_maxf=violin,
                    max_points=points,
                    zoom=zoom_active
                )

            self.components['graph']['feature_selector'].figure['layout'].clickmode = 'event+select'
            # Adjust graph with adding x and y axis titles
            self.components['graph']['feature_selector'].adjust_graph(
                x_ax=truncate_str(self.selected_feature, 110),
                y_ax='Contribution')
            return self.components['graph']['feature_selector'].figure

        @app.callback(
            [
                Output('index_id', 'value'),
                Output("index_id", "n_submit")
            ],
            [
                Input('feature_selector', 'clickData'),
                Input('prediction_picking', 'clickData'),
                Input('dataset', 'active_cell'),
                Input('apply_filter', 'n_clicks'),
                Input('reset_dropdown_button', 'n_clicks'),
                Input({'type': 'del_dropdown_button', 'index': ALL}, 'n_clicks')
            ],
            [
                State('dataset', 'data'),
                State('index_id', 'value')  # Get the current value of the index
            ]
        )
        def update_index_id(click_data,
                            prediction_picking,
                            cell,
                            apply_filters,
                            reset_filter,
                            nclicks_del,
                            data,
                            current_index_id):
            """
            This function is used to update index value according to
            active cell, filters and click data on feature plot or on
            prediction picking graph.
            ----------------------------------------------------------------
            click_data: click on feature selector
            prediction_picking: click on prediction picking graph
            cell: selected sell on dataset
            apply_filters: click on Apply filter button
            reset_filter: click on reset filter button
            nclicks_del: click on del button
            data: dataset
            current_index_id: the current value of the index
            ----------------------------------------------------------------
            return
            selected index id
            boolean n_submit
            """
            ctx = dash.callback_context
            selected = None
            if ctx.triggered[0]['prop_id'] != 'dataset.data':
                if ctx.triggered[0]['prop_id'] == 'feature_selector.clickData':
                    selected = click_data['points'][0]['customdata'][1]
                    self.click_graph = True
                elif ctx.triggered[0]['prop_id'] == 'prediction_picking.clickData':
                    selected = prediction_picking['points'][0]['customdata']
                    self.click_graph = True
                elif ctx.triggered[0]['prop_id'] == 'dataset.active_cell':
                    if cell is not None:
                        selected = data[cell['row']]['_index_']
                    else:
                        # Get actual value in field to refresh the selected value
                        selected = current_index_id
                elif ctx.triggered[0]['prop_id'] == '.':
                    selected = data[0]['_index_']
                # If click on Reset apply button
                elif ctx.triggered[0]['prop_id'] == 'reset_dropdown_button.n_clicks':
                    # Get the row index value
                    selected = data[cell['row']]['_index_']
                # If click on Apply filter button
                elif ctx.triggered[0]['prop_id'] == 'apply_filter.n_clicks':
                    # get the first index on the dataset
                    selected = data[0]['_index_']
                # If click on the last del button
                elif (('del_dropdown_button' in ctx.triggered[0]['prop_id']) &
                      (None not in nclicks_del)):
                    # Get the row index value
                    selected = data[cell['row']]['_index_']
                else:
                    selected = data[0]['_index_']
            else:
                raise PreventUpdate
            return selected, True

        @app.callback(
            Output('threshold_label', 'children'),
            [Input('threshold_id', 'value')])
        def update_threshold_label(value):
            """
            update threshold label
            """
            return f'Threshold: {value}'

        @app.callback(
            Output('max_contrib_label', 'children'),
            [Input('max_contrib_id', 'value')])
        def update_max_contrib_label(value):
            """
            update max_contrib label
            """
            self.components['filter']['max_contrib']['max_contrib_id'].value = value
            return f'Features to display: {value}'

        @app.callback(
            [Output('max_contrib_id', 'value'),
             Output('max_contrib_id', 'max'),
             Output('max_contrib_id', 'marks')
             ],
            [Input('modal', 'is_open')],
            [State('features', 'value')]
        )
        def update_max_contrib_id(is_open,
                                  features):
            """
            update max contrib component layout after settings modifications
            """
            ctx = dash.callback_context
            if ctx.triggered[0]['prop_id'] == 'modal.is_open':
                if is_open:
                    raise PreventUpdate
                else:
                    max = min(features, len(self.dataframe.columns) - 2)
                    if max // 5 == max / 5:
                        nb_marks = min(int(max // 5), 10)
                    elif max // 4 == max / 4:
                        nb_marks = min(int(max // 4), 10)
                    elif max // 3 == max / 3:
                        nb_marks = min(int(max // 3), 10)
                    elif max // 7 == max / 7:
                        nb_marks = min(int(max // 6), 10)
                    else:
                        nb_marks = 2
                    marks = {f'{round(max * feat / nb_marks)}': f'{round(max * feat / nb_marks)}'
                             for feat in range(1, nb_marks + 1)}
                    marks['1'] = '1'
                    if max < self.components['filter']['max_contrib']['max_contrib_id'].value:
                        value = max
                    else:
                        value = no_update

                    return value, max, marks

        @app.callback(
            Output(component_id='detail_feature', component_property='figure'),
            [
                Input('threshold_id', 'value'),
                Input('max_contrib_id', 'value'),
                Input('check_id_positive', 'value'),
                Input('check_id_negative', 'value'),
                Input('masked_contrib_id', 'value'),
                Input('select_label', 'value'),
                Input('dataset', 'active_cell'),
                Input('feature_selector', 'clickData'),
                Input('prediction_picking', 'clickData'),
                Input("validation", "n_clicks"),
                Input('bool_groups', 'on'),
                Input('ember_detail_feature', 'n_clicks'),
            ],
            [
                State('index_id', 'value'),
                State('dataset', 'data')
            ]
        )
        def update_detail_feature(threshold,
                                  max_contrib,
                                  positive,
                                  negative,
                                  masked,
                                  label,
                                  cell,
                                  click_data,
                                  prediction_picking,
                                  validation_click,
                                  bool_group,
                                  click_zoom,
                                  index,
                                  data):
            """
            update local explanation plot according to app changes.
            -------------------------------------------------------
            threshold: threshold
            max_contrib: max contribution
            positive: boolean
            negative: boolean
            masked: feature(s) to mask
            label: label
            cell: selected cell
            click_data: click on feature selector graph
            prediction_picking: click on prediction picking graph
            validation_click: click on validation
            bool_group: boolean
            click_zoom: click on zoom button
            index: selected index
            data: the dataset
            --------------------------------------------------------
            return
            detail feature graph
            """
            # Zoom is False by Default. It becomes True if we click on it
            click = 2 if click_zoom is None else click_zoom
            if click % 2 == 0:
                zoom_active = False
            else:
                zoom_active = True
            ctx = dash.callback_context
            selected = None
            if ctx.triggered[0]['prop_id'] == 'feature_selector.clickData':
                selected = click_data['points'][0]['customdata'][1]
            elif ctx.triggered[0]['prop_id'] == 'prediction_picking.clickData':
                selected = prediction_picking['points'][0]['customdata']
            elif ctx.triggered[0]['prop_id'] in ['threshold_id.value', 'validation.n_clicks']:
                selected = index
            elif ctx.triggered[0]['prop_id'] == 'dataset.active_cell':
                if cell:
                    selected = data[cell['row']]['_index_']
                else:
                    zoom_active = zoom_active
                    # raise PreventUpdate
            else:
                selected = index
            threshold = threshold if threshold != 0 else None
            if positive == [1]:
                sign = (None if negative == [1] else True)
            else:
                sign = (False if negative == [1] else None)
            self.explainer.filter(threshold=threshold,
                                  features_to_hide=masked,
                                  positive=sign,
                                  max_contrib=max_contrib,
                                  display_groups=bool_group)
            if np.issubdtype(type(self.explainer.x_init.index[0]), np.dtype(int).type):
                selected = int(selected)
            self.components['graph']['detail_feature'].figure = self.explainer.plot.local_plot(
                index=selected,
                label=label,
                show_masked=True,
                yaxis_max_label=8,
                display_groups=bool_group,
                zoom=zoom_active
            )
            # Adjust graph with adding x axis titles
            self.components['graph']['detail_feature'].adjust_graph(x_ax='Contribution')
            # font size can be adapted to screen size
            list_yaxis = [self.components['graph']['detail_feature'].figure.data[i].y[0] for i in
                          range(len(self.components['graph']['detail_feature'].figure.data))]
            # exclude new line with labels of y axis
            list_yaxis = [x.split('<br />')[0] for x in list_yaxis]
            nb_car = max([len(x) for x in list_yaxis])
            self.components['graph']['detail_feature'].figure.update_layout(
                yaxis=dict(tickfont={'size': min(round(500 / nb_car), 12)})
            )
            return self.components['graph']['detail_feature'].figure

        @app.callback(
            Output("validation", "n_clicks"),
            [
                Input("index_id", "n_submit")
            ],
        )
        def click_validation(n_submit):
            """
            submit index selection
            """
            if n_submit:
                return 1
            else:
                raise PreventUpdate

        @app.callback(
            [
                Output('dataset', 'style_data_conditional'),
                Output('dataset', 'style_filter_conditional'),
                Output('dataset', 'style_header_conditional'),
                Output('dataset', 'style_cell_conditional'),
            ],
            [
                Input("validation", "n_clicks")
            ],
            [
                State('dataset', 'data'),
                State('index_id', 'value')
            ]
        )
        def datatable_layout(validation,
                             data,
                             index):
            ctx = dash.callback_context
            if ctx.triggered[0]['prop_id'] == 'validation.n_clicks' and validation is not None:
                pass
            else:
                raise PreventUpdate

            style_data_conditional = [
                {
                    'if': {'row_index': 'odd'},
                    'backgroundColor': 'rgb(248, 248, 248)'
                }
            ]
            style_filter_conditional = []
            style_header_conditional = [
                {'if': {'column_id': c}, 'fontWeight': 'bold'}
                for c in ['_index_', '_predict_']
            ]
            style_cell_conditional = [
                {'if': {'column_id': c},
                 'width': '70px', 'fontWeight': 'bold'} for c in ['_index_', '_predict_']
            ]

            selected = check_row(data, index)
            if selected is not None:
                style_data_conditional += [{"if": {"row_index": selected}, "backgroundColor": self.color[0]}]

            return style_data_conditional, style_filter_conditional, style_header_conditional, style_cell_conditional

        @app.callback(
            Output(component_id='prediction_picking', component_property='figure'),
            [
                Input('global_feature_importance', 'clickData'),
                Input('dataset', 'data'),
                Input('apply_filter', 'n_clicks'),
                Input('reset_dropdown_button', 'n_clicks'),
                Input({'type': 'del_dropdown_button', 'index': ALL}, 'n_clicks'),
                Input('select_label', 'value'),
                Input('modal', 'is_open'),
                Input('ember_prediction_picking', 'n_clicks')
            ],
            [
                State('points', 'value'),
                State('violin', 'value')
            ]
        )
        def update_prediction_picking(feature,
                                      data,
                                      apply_filters,
                                      reset_filter,
                                      nclicks_del,
                                      label,
                                      is_open,
                                      click_zoom,
                                      points,
                                      violin):
            """
            Update feature plot according to label, data,
            selected feature and settings modifications
            ------------------------------------------------
            feature: click on features importance graph
            data: the dataset
            apply_filters: click on apply filter button
            reset_filter: click on reset filter button
            nclicks_del: click on del button
            label: selected label
            is_open: modal
            click_zoom: click on zoom button
            points: number of points
            violin: number of violin plot
            -------------------------------------------------
            return
            prediction picking graph
            """
            ctx = dash.callback_context
            # Filter subset
            filter_subset = None
            if not ctx.triggered:
                raise dash.exceptions.PreventUpdate
            if ctx.triggered[0]['prop_id'] == 'modal.is_open':
                if is_open:
                    raise PreventUpdate
                else:
                    self.settings['points'] = points
                    self.settings_ini['points'] = self.settings['points']
                    self.settings['violin'] = violin
                    self.settings_ini['violin'] = self.settings['violin']
            elif ctx.triggered[0]['prop_id'] == 'select_label.value':
                self.label = label
                self.subset = None
            # If we have clicked on reset button
            elif ctx.triggered[0]['prop_id'] == 'reset_dropdown_button.n_clicks':
                self.subset = None
            # If we have clicked on Apply filter button
            elif ctx.triggered[0]['prop_id'] == 'apply_filter.n_clicks':
                self.subset = [d['_index_'] for d in data]
            # If we have clicked on the last delete button (X)
            elif (('del_dropdown_button' in ctx.triggered[0]['prop_id']) &
                  (None not in nclicks_del)):
                self.subset = None
            else:
                raise PreventUpdate

            if self.explainer.y_target is not None:
                self.components['graph']['prediction_picking'].figure = self.explainer.plot.scatter_plot_prediction(
                    selection=self.subset,
                    max_points=points,
                    label=self.label
                )

                self.components['graph']['prediction_picking'].figure['layout'].clickmode = 'event+select'
                # Adjust graph with adding x and y axis titles
                self.components['graph']['prediction_picking'].adjust_graph(
                    x_ax="True Values",
                    y_ax="Predicted Values")
            else:
                fig = go.Figure()
                fig.update_layout(
                xaxis =  { "visible": False },
                yaxis = { "visible": False },
                annotations = [
                    {
                        "text": "Provide the y_target argument in the compile() method to display this plot.",
                        "xref": "paper",
                        "yref": "paper",
                        "showarrow": False,
                        "font": {
                            "size": 14
                        }
                    }
                ]
            )
                self.components['graph']['prediction_picking'].figure = fig

            return self.components['graph']['prediction_picking'].figure

        @app.callback(
            Output("modal_feature_importance", "is_open"),
            [Input("open_feature_importance", "n_clicks"),
             Input("close_feature_importance", "n_clicks")],
            [State("modal_feature_importance", "is_open")],
        )
        def toggle_modal_feature_importancet(n1,
                                             n2,
                                             is_open):
            """
            Function used to open and close modal explication when we click
            on "?" button on feature_importance graph
            ---------------------------------------------------------------
            n1: click on "?" button
            n2: click on close button in modal
            ---------------------------------------------------------------
            return modal
            """
            if n1 or n2:
                return not is_open
            return is_open

        @app.callback(
            Output("modal_feature_selector", "is_open"),
            [Input("open_feature_selector", "n_clicks"),
             Input("close_feature_selector", "n_clicks")],
            [State("modal_feature_selector", "is_open")],
        )
        def toggle_modal_feature_selector(n1,
                                          n2,
                                          is_open):
            """
            Function used to open and close modal explication when we click
            on "?" button on feature_selector graph
            ---------------------------------------------------------------
            n1: click on "?" button
            n2: click on close button in modal
            ---------------------------------------------------------------
            return modal
            """
            if n1 or n2:
                return not is_open
            return is_open

        @app.callback(
            Output("modal_detail_feature", "is_open"),
            [Input("open_detail_feature", "n_clicks"),
             Input("close_detail_feature", "n_clicks")],
            [State("modal_detail_feature", "is_open")],
        )
        def toggle_modal_detail_feature(n1,
                                        n2,
                                        is_open):
            """
            Function used to open and close modal explication when we click
            on "?" button on detail_feature graph
            ---------------------------------------------------------------
            n1: click on "?" button
            n2: click on close button in modal
            ---------------------------------------------------------------
            return modal
            """
            if n1 or n2:
                return not is_open
            return is_open

        @app.callback(
            Output("modal_prediction_picking", "is_open"),
            [Input("open_prediction_picking", "n_clicks"),
             Input("close_prediction_picking", "n_clicks")],
            [State("modal_prediction_picking", "is_open")],
        )
        def toggle_modal_prediction_picking(n1,
                                            n2,
                                            is_open):
            """
            Function used to open and close modal explication when we click
            on "?" button on prediction_picking graph
            ---------------------------------------------------------------
            n1: click on "?" button
            n2: click on close button in modal
            ---------------------------------------------------------------
            return modal
            """
            if n1 or n2:
                return not is_open
            return is_open

        @app.callback(
            Output("modal_filter", "is_open"),
            [Input("open_filter", "n_clicks"),
             Input("close_filter", "n_clicks")],
            [State("modal_filter", "is_open")],
        )
        def toggle_modal_filters(n1,
                                 n2,
                                 is_open):
            """
            Function used to open and close modal explication when we click
            on "?" button on Dataset Filters Tab
            ---------------------------------------------------------------
            n1: click on "?" button
            n2: click on close button in modal
            ---------------------------------------------------------------
            return modal
            """
            if n1 or n2:
                return not is_open
            return is_open

        # Add or remove plot blocs in the 'dropdowns_container'
        @app.callback(
            Output('dropdowns_container', 'children'),
            [
                Input('add_dropdown_button', 'n_clicks'),
                Input('reset_dropdown_button', 'n_clicks'),
                Input({'type': 'del_dropdown_button', 'index': ALL}, 'n_clicks')
            ],
            [
                 State('dropdowns_container', 'children'),
                 State('name', 'value')
            ]
        )
        def layout_filter(n_clicks_add,
                          n_clicks_rm,
                          n_clicks_reset,
                          currents_filters,
                          name
                          ):
            """
            Function used to create filter blocs in the dropdowns_container.
            Each bloc will contains:
                -label
                -dropdown button to select feature to filter
                -div which will contains modalities
                -delete button
            ---------------------------------------------------------------
            n_clicks_add: click on add filter
            n_clicks_reset: click on reset filter button
            n_click_del: click on delete button
            children: information on dropdown container
            name: name for feature name
            ---------------------------------------------------------------
            return
                filter blocs
            """
            # Context and init handling (no action)
            ctx = dash.callback_context
            if not ctx.triggered:
                raise dash.exceptions.PreventUpdate
            button_id = ctx.triggered[0]['prop_id'].split('.')[0]

            # We use domain name for feature name
            dict_name = [self.explainer.features_dict[i]
                         for i in self.dataframe.drop(['_index_', '_predict_'], axis=1).columns]
            dict_id = [i for i in self.dataframe.drop(['_index_', '_predict_'], axis=1).columns]
            # Create dataframe to sort it by feature_name
            df_feature_name = pd.DataFrame({'feature_name': dict_name,
                                            'feature_id': dict_id})
            df_feature_name = df_feature_name.sort_values(
                by='feature_name').reset_index(drop=True)
            # Options are sorted by feature_name
            options = [
                {"label": '_index_', "value": '_index_'},
                {"label": '_predict_', "value": '_predict_'}] + \
                [{"label": df_feature_name.loc[i, 'feature_name'],
                  "value": df_feature_name.loc[i, 'feature_id']}
                 for i in range(len(df_feature_name))]

            # Creation of a new graph
            if button_id == 'add_dropdown_button':
                # ID index definition
                if n_clicks_add is None:
                    index_id = 0
                else:
                    index_id = n_clicks_add
                # Appending a dropdown block to 'dropdowns_container'children
                subset_filter = html.Div(
                    id={'type': 'bloc_div',
                        'index': index_id},
                    children=[
                        html.Div([
                            html.Br(),
                            # div which will contains label
                            html.Div(
                                    id={'type': 'dynamic-output-label',
                                        'index': index_id},
                                    )
                            ]),
                        html.Div([
                            # div with dopdown button to select feature to filter
                            html.Div(dcc.Dropdown(
                                id={'type': 'var_dropdown',
                                    'index': index_id},
                                options=options,
                                placeholder="Variable"
                            ), style={"width": "30%"}),
                            # div which will contains modalities
                            html.Div(
                                 id={'type': 'dynamic-output',
                                     'index': index_id},
                                 style={"width": "50%"}
                                ),
                            # Button to delete bloc
                            dbc.Button(
                                id={'type': 'del_dropdown_button',
                                    'index': index_id},
                                children='X',
                                color='warning',
                                size='sm'
                            )
                        ], style={'display': 'flex'})
                    ]
                )
                return currents_filters + [subset_filter]
            # Removal of all existing filters
            elif button_id == 'reset_dropdown_button':
                return [html.Div(
                    id={'type': 'bloc_div',
                        'index': 0},
                    children=[])]
            # Removal of an existing filter
            else:
                filter_id_to_remove = eval(button_id)['index']
                return [gr for gr in currents_filters
                        if gr['props']['id']['index'] != filter_id_to_remove]

        @app.callback(
            Output('reset_dropdown_button', 'disabled'),
            [Input('add_dropdown_button', 'n_clicks'),
             Input('reset_dropdown_button', 'n_clicks'),
             Input({'type': 'del_dropdown_button', 'index': ALL}, 'n_clicks')]
        )
        def update_disabled_reset_button(n_click_add,
                                         n_click_reset,
                                         n_click_del):
            """
            Function used to disabled or not the reset filter button.
            This button is disabled if there is no filter added.
            ---------------------------------------------------------------
            n_click_add: click on add filter button
            n_click_reset: click on reset filter button
            n_click_del: click on delete button
            ---------------------------------------------------------------
            return disabled style
            """
            ctx = dash.callback_context
            if ctx.triggered[0]['prop_id'] == 'add_dropdown_button.n_clicks':
                disabled = False
            elif ctx.triggered[0]['prop_id'] == 'reset_dropdown_button.n_clicks':
                disabled = True
            elif None not in n_click_del:
                disabled = True
            else:
                disabled = False
            return disabled

        @app.callback(
            Output('apply_filter', 'style'),
            [Input('add_dropdown_button', 'n_clicks'),
             Input('reset_dropdown_button', 'n_clicks'),
             Input({'type': 'del_dropdown_button', 'index': ALL}, 'n_clicks')]
        )
        def update_style_apply_filter_button(n_click_add,
                                             n_click_reset,
                                             n_click_del):
            """
            Function used to display or not the apply filter button.
            This button is only display if almost one filter was added.
            ---------------------------------------------------------------
            n_click_add: click on add filter button
            n_click_reset: click on reset filter button
            n_click_del: click on delete button
            ---------------------------------------------------------------
            return style of apply filter button
            """
            ctx = dash.callback_context
            if ctx.triggered[0]['prop_id'] == 'add_dropdown_button.n_clicks':
                return {'display': 'block'}
            elif ctx.triggered[0]['prop_id'] == 'reset_dropdown_button.n_clicks':
                return {'display': 'none'}
            elif None not in n_click_del:
                return {'display': 'none'}
            else:
                return {'display': 'block'}

        @app.callback(
            Output({'type': 'dynamic-output-label', 'index': MATCH}, 'children'),
            Input({'type': 'var_dropdown', 'index': MATCH}, 'value'),
        )
        def update_label_filter(value):
            """
            Function used to add label to the filters. Label is updated
            when value is not None
            ---------------------------------------------------------------
            value: value selected on the var dropdown button
            ---------------------------------------------------------------
            return label
            """
            if value is not None:
                return html.Label("Variable {} is filtered".format(value))
            else:
                return html.Label('Select variable to filter')

        @app.callback(
            Output({'type': 'dynamic-output', 'index': MATCH}, 'children'),
            [Input({'type': 'var_dropdown', 'index': MATCH}, 'value'),
             Input({'type': 'var_dropdown', 'index': MATCH}, 'id'),
             Input('add_dropdown_button', 'n_clicks')],
        )
        def display_output(value,
                           id,
                           add_click):
            """
            Function used to create modalities choices. Componenents are different
            according to the type of the selected variable.
            For string variable: component is a dropdown button
            For boolean variable: component is a RadioItems button
            For Integer variable that have less than 20 modalities: component
            is a dropdown button.
            For date variable: component is a DatePickerRange
            Else: components are lower and upper values
            ---------------------------------------------------------------
            value: value selected on the var dropdown button
            id: id of the var dropdown button
            add_click: click on add_dropdown_button
            ---------------------------------------------------------------
            return modalities components. If the component is new, value
            is empty by default.
            """
            # Context and init handling (no action)
            ctx = dash.callback_context
            if not ctx.triggered:
                raise dash.exceptions.PreventUpdate
            # No update last modalities values if we click on add button
            if ctx.triggered[0]['prop_id'] == 'add_dropdown_button.n_clicks':
                raise dash.exceptions.PreventUpdate
            # Creation on modalities dropdown button
            else:
                if value is not None:
                    if type(self.round_dataframe[value].iloc[0]) == bool:
                        new_element = html.Div(dcc.RadioItems(
                            [{'label': val, 'value': val} for
                             val in self.round_dataframe[value].unique()],
                            id={'type': 'dynamic-bool',
                                'index': id['index']},
                            value=self.round_dataframe[value].iloc[0],
                            inline=False
                            ), style={"width": "65%", 'margin-left': '20px'})
                    elif (type(self.round_dataframe[value].iloc[0]) == str) | \
                         ((type(self.round_dataframe[value].iloc[0]) == np.int64) &
                          (len(self.round_dataframe[value].unique()) <= 20)):
                        new_element = html.Div(dcc.Dropdown(
                            id={
                               'type': 'dynamic-str',
                               'index': id['index']
                            },
                            options=[{'label': i, 'value': i} for
                                    i in np.sort(self.round_dataframe[value].unique())],
                            multi=True,
                            ), style={"width": "65%", 'margin-left': '20px'})
                    elif ((type(self.round_dataframe[value].iloc[0]) is pd.Timestamp) |
                          (type(self.round_dataframe[value].iloc[0]) is datetime.datetime)):
                        new_element = html.Div(
                            dcc.DatePickerRange(
                                id={
                                   'type': 'dynamic-date',
                                   'index': id['index']
                                },
                                min_date_allowed=self.round_dataframe[value].min(),
                                max_date_allowed=self.round_dataframe[value].max(),
                                start_date=self.round_dataframe[value].min(),
                                end_date=self.round_dataframe[value].max()
                               ), style={'width': '65%', 'margin-left': '20px'}),
                    else:
                        lower_value = 0
                        upper_value = 0
                        new_element = html.Div([
                                            dcc.Input(
                                                id={
                                                    'type': 'lower',
                                                    'index': id['index']
                                                },
                                                value=lower_value,
                                                type="number",
                                                style={'width': '60px'}),
                                            ' <= {} in [{}, {}]<= '.format(
                                                value,
                                                self.round_dataframe[value].min(),
                                                self.round_dataframe[value].max()),
                                            dcc.Input(
                                                id={
                                                    'type': 'upper',
                                                    'index': id['index']
                                                },
                                                value=upper_value,
                                                type="number",
                                                style={'width': '60px'}
                                            )
                        ], style={'margin-left': '20px'})
                else:
                    new_element = html.Div()
                return new_element



"""
Class inherited from dcc.Graph. Add one method for updating graph layout.
"""

from dash import dcc
import re


class MyGraph(dcc.Graph):

    def __init__(self, figure, id, style={}, **kwds):
        super().__init__(**kwds)
        self.figure = figure
        self.style = style
        self.id = id
        # self.config = {'modeBarButtons': {'pan2d': True}}
        if id == 'prediction_picking':
            self.config = {
                # Graph is responsive
                'responsive': True,
                'modeBarButtonsToRemove': ['zoomOut2d',
                                           'zoomIn2d',
                                           'resetScale2d',
                                           'hoverClosestCartesian',
                                           'hoverCompareCartesian',
                                           'toggleSpikelines'],
                'displaylogo': False
                }
            # 'modeBarStyle': {'orientation': 'v'}, # Deprecated in Dash 1.17.0
        else:
            self.config = {
                # Graph is responsive
                'responsive': True,
                # Graph don't have select box button
                'modeBarButtonsToRemove': ['lasso2d',
                                           'zoomOut2d',
                                           'zoomIn2d',
                                           'resetScale2d',
                                           'hoverClosestCartesian',
                                           'hoverCompareCartesian',
                                           'toggleSpikelines',
                                           'select'],
                'displaylogo': False
                }

    def adjust_graph(self,
                     x_ax="",
                     y_ax=""):
        """
        Override graph layout for app use
        ----------------------------------------
        x_ax: title of the x-axis
        y_ax: title of the y-axis
        ---------------------------------------
        """
        new_title = update_title(self.figure.layout.title.text)
        self.figure.update_layout(
            autosize=True,
            margin=dict(
                l=50,
                r=10,
                b=10,
                t=67,
                pad=0
            ),
            width=None,
            height=None,
            title={
                'y': 0.95,
                'x': 0.5,
                'xanchor': 'center',
                'yanchor': 'top',
                # update title and font-size of the title
                'text': '<span style="font-size: 1.2vw;">' + new_title + '</span>'
            }
        )
        # update x title and font-size of the title
        self.figure.update_xaxes(title='<span style="font-size: 1vw;">' + x_ax + '</span>',
                                 automargin=True
                                 )
        # update y title and font-size of the title
        self.figure.update_yaxes(title='<span style="font-size: 1vw;">' + y_ax + '</span>',
                                 automargin=True
                                 )


def update_title(title):
    """
    adapt title content the app layout
    Parameters
    ----------
    title : str
        string to ba adapted
    Returns
    -------
    str
    """
    patt = re.compile('^(.+)<span.+?(Predict: .*|Proba: .*)?</span>$')
    try:
        list_non_empty_str_matches = [x for x in patt.findall(title)[0] if x != '']
        updated = ' - '.join(map(str, list_non_empty_str_matches))
    except:
        updated = title
    return updated



"""
Data loader module
"""
import os
import json
import pandas as pd
from urllib.request import urlretrieve


def data_loading(dataset):
    """
    data_loading allows shapash user to try the library with small but clear datasets.
    Titanic, house_prices or telco_customer_churn data.

    Example
    ----------
    >>> from shapash.data.data_loader import data_loading
    >>> house_df, house_dict = data_loading('house_prices')

    Parameters
    ----------
    dataset : String
        Dataset's name to return.
         - 'titanic'
         - 'house_prices'
         - 'telco_customer_churn'

    Returns
    -------
    data : pandas.DataFrame
        Dataset required
    dict : (Dictionnary, Optional)
        If exist, columns labels dictionnary associated to the dataset.
    """
    current_path = os.path.dirname(os.path.abspath(__file__))
    if dataset == 'house_prices':
        if os.path.isfile(current_path+'/house_prices_dataset.csv') is False:
            github_data_url = 'https://github.com/MAIF/shapash/raw/master/shapash/data/'
            urlretrieve(github_data_url + "house_prices_dataset.csv", filename=current_path + "/house_prices_dataset.csv")
            urlretrieve(github_data_url + "house_prices_labels.json", filename=current_path + "/house_prices_labels.json")
        data_house_prices_path = os.path.join(current_path, "house_prices_dataset.csv")
        dict_house_prices_path = os.path.join(current_path, "house_prices_labels.json")
        data = pd.read_csv(data_house_prices_path, header=0, index_col=0, engine='python')
        with open(dict_house_prices_path, 'r') as openfile2:
            dic = json.load(openfile2)
        return data, dic

    elif dataset == 'titanic':
        if os.path.isfile(current_path + '/titanicdata.csv') is False:
            github_data_url = 'https://github.com/MAIF/shapash/raw/master/shapash/data/'
            urlretrieve(github_data_url + "titanicdata.csv", filename=current_path + "/titanicdata.csv")
            urlretrieve(github_data_url + "titaniclabels.json", filename=current_path + "/titaniclabels.json")
        data_titanic_path = os.path.join(current_path, "titanicdata.csv")
        dict_titanic_path = os.path.join(current_path, 'titaniclabels.json')
        data = pd.read_csv(data_titanic_path, header=0, index_col=0, engine='python')
        with open(dict_titanic_path, 'r') as openfile:
            dic = json.load(openfile)
        return data, dic

    elif dataset == 'telco_customer_churn':
        if os.path.isfile(current_path + '/telco_customer_churn.csv') is False:
            github_data_url = 'https://github.com/IBM/telco-customer-churn-on-icp4d/raw/master/data/'
            urlretrieve(github_data_url + "Telco-Customer-Churn.csv", filename=current_path + "/telco_customer_churn.csv")
        data_telco_path = os.path.join(current_path, "telco_customer_churn.csv")
        data = pd.read_csv(data_telco_path, header=0, index_col=0, engine='python')
        return data

    else:
        raise ValueError("Dataset not found. Check the docstring for available values")



from typing import Union, Optional
import pandas as pd
import numpy as np
import seaborn as sns
import matplotlib.pyplot as plt
from matplotlib.colors import LinearSegmentedColormap

from shapash.utils.utils import truncate_str
from shapash.report.common import VarType
from shapash.style.style_utils import get_pyplot_color, get_palette


def generate_fig_univariate(
        df_all: pd.DataFrame,
        col: str,
        hue: str,
        type: VarType,
        colors_dict: Optional[dict] = None
) -> plt.Figure:
    """
    Returns a matplotlib figure containing the distribution of any kind of feature
    (continuous, categorical).

    If the feature is categorical and contains too many categories, the smallest
    categories are grouped into a new 'Other' category so that the graph remains
    readable.

    The input dataframe should contain the column of interest and a column that is used
    to distinguish two types of values (ex. 'train' and 'test')

    Parameters
    ----------
    df_all : pd.DataFrame
        The input dataframe that contains the column of interest
    col : str
        The column of interest
    hue : str
        The column used to distinguish the values (ex. 'train' and 'test')
    type: str
        The type of the series ('continous' or 'categorical')
    colors_dict : dict
        dict of colors used

    Returns
    -------
    matplotlib.pyplot.Figure
    """
    if type == VarType.TYPE_NUM:
        fig = generate_fig_univariate_continuous(df_all, col, hue=hue, colors_dict=colors_dict)
    elif type == VarType.TYPE_CAT:
        fig = generate_fig_univariate_categorical(df_all, col, hue=hue, colors_dict=colors_dict)
    else:
        raise NotImplemented("Series dtype not supported")
    return fig


def generate_fig_univariate_continuous(
        df_all: pd.DataFrame,
        col: str,
        hue: str,
        colors_dict: Optional[dict] = None
) -> plt.Figure:
    """
    Returns a matplotlib figure containing the distribution of a continuous feature.

    Parameters
    ----------
    df_all : pd.DataFrame
        The input dataframe that contains the column of interest
    col : str
        The column of interest
    hue : str
        The column used to distinguish the values (ex. 'train' and 'test')
    colors_dict : dict
        dict of colors used

    Returns
    -------
    matplotlib.pyplot.Figure
    """
    colors_dict = colors_dict or get_palette('default')
    g = sns.displot(df_all, x=col, hue=hue, kind="kde", fill=True, common_norm=False,
                    palette=get_pyplot_color(colors=colors_dict['report_feature_distribution']))
    g.set_xticklabels(rotation=30)

    fig = g.fig

    fig.set_figwidth(7)
    fig.set_figheight(4)

    return fig


def generate_fig_univariate_categorical(
        df_all: pd.DataFrame,
        col: str,
        hue: str,
        nb_cat_max: int = 7,
        colors_dict: Optional[dict] = None
) -> plt.Figure:
    """
    Returns a matplotlib figure containing the distribution of a categorical feature.

    If the feature is categorical and contains too many categories, the smallest
    categories are grouped into a new 'Other' category so that the graph remains
    readable.

    Parameters
    ----------
    df_all : pd.DataFrame
        The input dataframe that contains the column of interest
    col : str
        The column of interest
    hue : str
        The column used to distinguish the values (ex. 'train' and 'test')
    nb_cat_max : int
        The number max of categories to be displayed. If the number of categories
        is greater than nb_cat_max then groups smallest categories into a new
        'Other' category
    colors_dict : dict
        dict of colors used

    Returns
    -------
    matplotlib.pyplot.Figure
    """
    colors_dict = colors_dict or get_palette('default')
    df_cat = df_all.groupby([col, hue]).agg({col: 'count'})\
                   .rename(columns={col: "count"}).reset_index()
    df_cat['Percent'] = df_cat['count'] * 100 / df_cat.groupby(hue)['count'].transform('sum')

    if pd.api.types.is_numeric_dtype(df_cat[col].dtype):
        df_cat = df_cat.sort_values(col, ascending=True)
        df_cat[col] = df_cat[col].astype(str)

    nb_cat = df_cat.groupby([col]).agg({'count': 'sum'}).reset_index()[col].nunique()

    if nb_cat > nb_cat_max:
        df_cat = _merge_small_categories(df_cat=df_cat, col=col, hue=hue, nb_cat_max=nb_cat_max)

    fig, ax = plt.subplots(figsize=(7, 4))

    sns.barplot(data=df_cat, x='Percent', y=col, hue=hue,
                palette=get_pyplot_color(colors=colors_dict['report_feature_distribution']), ax=ax)

    for p in ax.patches:
        ax.annotate("{:.1f}%".format(np.nan_to_num(p.get_width(), nan=0)),
                    xy=(p.get_width(), p.get_y() + p.get_height() / 2),
                    xytext=(5, 0), textcoords='offset points', ha="left", va="center")

    # Shrink current axis by 20%
    box = ax.get_position()
    ax.set_position([box.x0, box.y0, box.width * 0.8, box.height])

    # Put a legend to the right of the current axis
    ax.legend(loc='center left', bbox_to_anchor=(1, 0.5))

    # Removes plot borders
    ax.spines['top'].set_visible(False)
    ax.spines['right'].set_visible(False)

    new_labels = [truncate_str(i.get_text(), maxlen=45) for i in ax.yaxis.get_ticklabels()]
    ax.yaxis.set_ticklabels(new_labels)

    return fig


def _merge_small_categories(df_cat: pd.DataFrame, col: str, hue: str,  nb_cat_max: int) -> pd.DataFrame:
    """
    Merges categories of column 'col' of df_cat into 'Other' category so that
    the number of categories is less than nb_cat_max.
    """
    df_cat_sum_hue = df_cat.groupby([col]).agg({'count': 'sum'}).reset_index()
    list_cat_to_merge = df_cat_sum_hue.sort_values('count', ascending=False)[col].to_list()[nb_cat_max - 1:]
    df_cat_other = df_cat.loc[df_cat[col].isin(list_cat_to_merge)] \
        .groupby(hue, as_index=False)[["count", "Percent"]].sum()
    df_cat_other[col] = "Other"
    return df_cat.loc[~df_cat[col].isin(list_cat_to_merge)].append(df_cat_other)


def generate_confusion_matrix_plot(
        y_true: Union[np.array, list],
        y_pred: Union[np.array, list],
        colors_dict: Optional[dict] = None
) -> plt.Figure:
    """
    Returns a matplotlib figure containing a confusion matrix that is computed using y_true and
    y_pred parameters.

    Parameters
    ----------
    y_true : array-like
        Ground truth (correct) target values.
    y_pred : array-like
        Estimated targets as returned by a classifier.
    colors_dict : dict
        dict of colors used
    Returns
    -------
    matplotlib.pyplot.Figure
    """
    colors_dict = colors_dict or get_palette('default')
    col_scale = get_pyplot_color(colors=colors_dict['report_confusion_matrix'])
    cmap_gradient = LinearSegmentedColormap.from_list('col_corr', col_scale, N=100)

    df_cm = pd.crosstab(y_true, y_pred, rownames=['Actual'], colnames=['Predicted'])
    fig, ax = plt.subplots(figsize=(7, 4))
    sns.heatmap(df_cm, ax=ax, annot=True, cmap=cmap_gradient, fmt='g')
    return fig



from typing import Optional, Union, Tuple
import logging
import sys
import os
from numbers import Number
from datetime import date
import jinja2
import pandas as pd
import numpy as np
import plotly

from shapash.utils.transform import inverse_transform, apply_postprocessing
from shapash import SmartExplainer
from shapash.utils.io import load_yml
from shapash.utils.utils import get_project_root, truncate_str
from shapash.report.visualisation import print_md, print_html, print_css_style, convert_fig_to_html, \
    print_javascript_misc
from shapash.report.data_analysis import perform_global_dataframe_analysis, perform_univariate_dataframe_analysis
from shapash.report.plots import generate_fig_univariate, generate_confusion_matrix_plot
from shapash.report.common import series_dtype, get_callable, compute_col_types, VarType, display_value
from shapash.webapp.utils.utils import round_to_k

logging.basicConfig(level=logging.INFO)

template_loader = jinja2.FileSystemLoader(searchpath=os.path.join(get_project_root(), 'shapash', 'report', 'html'))
template_env = jinja2.Environment(loader=template_loader)


class ProjectReport:
    """
    The ProjectReport class allows to generate general information about a
    Data Science project.
    It analyzes the data and the model used in order to provide interesting
    insights that can be shared with non technical person.

    Parameters
    ----------
    explainer : shapash.explainer.smart_explainer.SmartExplainer
        A shapash SmartExplainer object that has already be compiled.
    project_info_file : str
        Path to the yml file containing information about the project (author, description, ...).
    config : dict, optional
        Contains configuration options for the report.

    Attributes
    ----------
    explainer : shapash.explainer.smart_explainer.SmartExplainer
         A shapash SmartExplainer object that has already be compiled.
    metadata : dict
        Information about the project (author, description, ...).
    x_train : pd.DataFrame
        DataFrame used for training the model.
    y_test : pd.Series or pd.DataFrame
        Series of labels in the train set.
    y_test : pd.Series or pd.DataFrame
        Series of labels in the test set.
    config : dict, optional
        Configuration options for the report.

    """

    def __init__(
            self,
            explainer: SmartExplainer,
            project_info_file: str,
            x_train: Optional[pd.DataFrame] = None,
            y_train: Optional[pd.DataFrame] = None,
            y_test: Optional[pd.DataFrame] = None,
            config: Optional[dict] = None
    ):
        self.explainer = explainer
        self.metadata = load_yml(path=project_info_file)
        self.x_train_init = x_train
        if x_train is not None:
            self.x_train_pre = inverse_transform(x_train, self.explainer.preprocessing)
            if self.explainer.postprocessing:
                self.x_train_pre = apply_postprocessing(self.x_train_pre, self.explainer.postprocessing)
        else:
            self.x_train_pre = None
        self.x_init = self.explainer.x_init
        self.config = config if config is not None else dict()
        self.col_names = list(self.explainer.columns_dict.values())
        self.df_train_test = self._create_train_test_df(test=self.x_init, train=self.x_train_pre)
        if self.explainer.y_pred is not None:
            self.y_pred = np.array(self.explainer.y_pred.T)[0]
        else:
            self.y_pred = self.explainer.model.predict(self.explainer.x_encoded)
        self.y_test, target_name_test = self._get_values_and_name(y_test, 'target')
        self.y_train, target_name_train = self._get_values_and_name(y_train, 'target')
        self.target_name = target_name_train or target_name_test

        if 'title_story' in self.config.keys():
            self.title_story = config['title_story']
        elif self.explainer.title_story != '':
            self.title_story = self.explainer.title_story
        else:
            self.title_story = 'Shapash report'
        self.title_description = self.config['title_description'] if 'title_description' in self.config.keys() else ''

        print_css_style()
        print_javascript_misc()

        if 'metrics' in self.config.keys():
            if not isinstance(self.config['metrics'], list) or not isinstance(self.config['metrics'][0], dict):
                raise ValueError("The metrics parameter expects a list of dict.")
            for metric in self.config['metrics']:
                for key in metric:
                    if key not in ['path', 'name', 'use_proba_values']:
                        raise ValueError(f"Unknown key : {key}. Key should be in ['path', 'name', 'use_proba_values']")
                    if key == 'use_proba_values' and not isinstance(metric['use_proba_values'], bool):
                        raise ValueError('"use_proba_values" metric key expects a boolean value.')

    @staticmethod
    def _get_values_and_name(
            y: Optional[Union[pd.DataFrame, pd.Series, list]],
            default_name: str
    ) -> Union[Tuple[list, str], Tuple[None, None]]:
        """
        Extracts vales and column name from a Pandas Series, DataFrame, or assign a default
        name if y is a list of values.

        Parameters
        ----------
        y : list or pd.Series or pd.DataFrame
            Column we want to extract the name and values
        default_name :
            Name assigned if no name was found for y

        Returns
        -------
        values : list
            list of values of y
        name : str
            name of y
        """
        if y is None:
            return None, None
        elif isinstance(y, pd.DataFrame):
            assert len(y.columns) == 1, "Number of columns found is greater than 1"
            name = y.columns[0]
            values = y.values[:, 0]
        elif isinstance(y, pd.Series):
            name = y.name
            values = y.values
        elif isinstance(y, list):
            name = default_name
            values = y
        else:
            raise ValueError(f"Cannot process following type : {type(y)}")
        return values, name

    @staticmethod
    def _create_train_test_df(test: Optional[pd.DataFrame], train: Optional[pd.DataFrame]) -> Union[pd.DataFrame, None]:
        """
        Creates a DataFrame that contains train and test dataset with the column 'data_train_test'
        allowing to distinguish the values.

        Parameters
        ----------
        test : pd.DataFrame, optional
            test dataframe
        train : pd.DataFrame, optional
            train dataframe

        Returns
        -------
        pd.DataFrame
            The concatenation of train and test as a dataframe containing train and test values with
            a new 'data_train_test' column allowing to distinguish the values.
        """
        if (test is not None and 'data_train_test' in test.columns) or \
                (train is not None and 'data_train_test' in train.columns):
            raise ValueError('"data_train_test" column must be renamed as it is used in ProjectReport')
        if test is None and train is None:
            return None
        return pd.concat([test.assign(data_train_test="test") if test is not None else None,
                          train.assign(data_train_test="train") if train is not None else None]).reset_index(drop=True)

    def display_title_description(self):
        """
        Displays title of the report and its description if defined.
        """
        print_html(f"""<h1 style="text-align:center">{self.title_story}</p> """)
        if self.title_description != '':
            print_html(f'<blockquote class="panel-warning text_cell_render">{self.title_description} </blockquote>')

    def display_project_information(self):
        """
        Displays general information about the project as defined in the metdata file.
        """
        for section in self.metadata.keys():
            print_md(f"## {section.title()}")
            for k, v in self.metadata[section].items():
                if k.lower() == 'date' and v.lower() == 'auto':
                    print_md(f"**{k.title()}** : {date.today()}")
                else:
                    print_md(f"**{k.title()}** : {v}")
            print_md('---')

    def display_model_analysis(self):
        """
        Displays information about the model used : class name, library name, library version,
        model parameters, ...
        """
        print_md(f"**Model used :** {self.explainer.model.__class__.__name__}")

        print_md(f"**Library :** {self.explainer.model.__class__.__module__}")

        for name, module in sorted(sys.modules.items()):
            if hasattr(module, '__version__') \
                    and self.explainer.model.__class__.__module__.split('.')[0] == module.__name__:
                print_md(f"**Library version :** {module.__version__}")

        print_md("**Model parameters :** ")
        model_params = self.explainer.model.__dict__
        table_template = template_env.get_template("double_table.html")
        print_html(table_template.render(
            columns1=["Parameter key", "Parameter value"],
            rows1=[{"name": truncate_str(str(k), 50), "value": truncate_str(str(v), 300)}
                   for k, v in list(model_params.items())[:len(model_params)//2:]],  # Getting half of the parameters
            columns2=["Parameter key", "Parameter value"],
            rows2=[{"name": truncate_str(str(k), 50), "value": truncate_str(str(v), 300)}
                   for k, v in list(model_params.items())[len(model_params)//2:]]  # Getting 2nd half of the parameters
        ))
        print_md('---')

    def display_dataset_analysis(
            self,
            global_analysis: bool = True,
            univariate_analysis: bool = True,
            target_analysis: bool = True,
            multivariate_analysis: bool = True
    ):
        """
        This method performs and displays an exploration of the data given.
        It allows to compare train and test values for each part of the analysis.

        The parameters of the method allow to filter which part to display or not.

        Parameters
        ----------
        global_analysis : bool
            Whether or not to display the global analysis part.
        univariate_analysis : bool
            Whether or not to display the univariate analysis part.
        target_analysis : bool
            Whether or not to display the target analysis part that plots
            the distribution of the target variable.
        multivariate_analysis : bool
            Whether or not to display the multivariate analysis part
        """
        if global_analysis:
            print_md("### Global analysis")
            self._display_dataset_analysis_global()

        if univariate_analysis:
            print_md("### Univariate analysis")
            self._perform_and_display_analysis_univariate(
                df=self.df_train_test,
                col_splitter="data_train_test",
                split_values=["test", "train"],
                names=["Prediction dataset", "Training dataset"],
                group_id='univariate'
            )
        if target_analysis:
            df_target = self._create_train_test_df(
                test=pd.DataFrame({self.target_name: self.y_test},
                                  index=range(len(self.y_test))) if self.y_test is not None else None,
                train=pd.DataFrame({self.target_name: self.y_train},
                                   index=range(len(self.y_train))) if self.y_train is not None else None
            )
            if df_target is not None:
                if target_analysis:
                    print_md("### Target analysis")
                    self._perform_and_display_analysis_univariate(
                        df=df_target,
                        col_splitter="data_train_test",
                        split_values=["test", "train"],
                        names=["Prediction dataset", "Training dataset"],
                        group_id='target'
                    )
        if multivariate_analysis:
            print_md("### Multivariate analysis")
            fig_corr = self.explainer.plot.correlations(
                self.df_train_test,
                facet_col='data_train_test',
                max_features=20,
                width=900 if len(self.df_train_test['data_train_test'].unique()) > 1 else 500,
                height=500,
            )
            print_html(plotly.io.to_html(fig_corr))
        print_md('---')

    def _display_dataset_analysis_global(self):
        df_stats_global = self._stats_to_table(test_stats=perform_global_dataframe_analysis(self.x_init),
                                               train_stats=perform_global_dataframe_analysis(self.x_train_pre),
                                               names=["Prediction dataset", "Training dataset"])
        print_html(df_stats_global.to_html(classes="greyGridTable"))

    def _perform_and_display_analysis_univariate(
            self,
            df: pd.DataFrame,
            col_splitter: str,
            split_values: list,
            names: list,
            group_id: str
    ):
        col_types = compute_col_types(df)
        n_splits = df[col_splitter].nunique()
        inv_columns_dict = {v: k for k, v in self.explainer.columns_dict.items()}
        test_stats_univariate = perform_univariate_dataframe_analysis(df.loc[df[col_splitter] == split_values[0]],
                                                                      col_types=col_types)
        if n_splits > 1:
            train_stats_univariate = perform_univariate_dataframe_analysis(df.loc[df[col_splitter] == split_values[1]],
                                                                           col_types=col_types)

        univariate_template = template_env.get_template("univariate.html")
        univariate_features_desc = list()
        list_cols_labels = [self.explainer.features_dict.get(col, col)
                            for col in df.drop(col_splitter, axis=1).columns.to_list()]
        for col_label in sorted(list_cols_labels):
            col = self.explainer.inv_features_dict.get(col_label, col_label)
            fig = generate_fig_univariate(df_all=df, col=col, hue=col_splitter, type=col_types[col],
                                          colors_dict=self.explainer.colors_dict)
            df_col_stats = self._stats_to_table(
                test_stats=test_stats_univariate[col],
                train_stats=train_stats_univariate[col] if n_splits > 1 else None,
                names=names
            )

            univariate_features_desc.append({
                'feature_index': int(inv_columns_dict.get(col, 0)),
                'name': col,
                'type': str(series_dtype(df[col])),
                'description': col_label,
                'table': df_col_stats.to_html(classes="greyGridTable"),
                'image': convert_fig_to_html(fig)
            })
        print_html(univariate_template.render(features=univariate_features_desc, groupId=group_id))

    @staticmethod
    def _stats_to_table(test_stats: dict,
                        names: list,
                        train_stats: Optional[dict] = None,
                        ) -> pd.DataFrame:
        if train_stats is not None:
            return pd.DataFrame({
                names[1]: pd.Series(train_stats),
                names[0]: pd.Series(test_stats)
            })
        else:
            return pd.DataFrame({names[0]: pd.Series(test_stats)})

    def display_model_explainability(self):
        """
        Displays explainability of the model as computed in SmartPlotter object
        """
        print_md("*Note : the explainability graphs were generated using the test set only.*")
        explainability_template = template_env.get_template("explainability.html")
        inv_columns_dict = {v: k for k, v in self.explainer.columns_dict.items()}
        explain_data = list()
        multiclass = True if (self.explainer._classes and len(self.explainer._classes) > 2) else False
        c_list = self.explainer._classes if multiclass else [1]  # list just used for multiclass
        for index_label, label in enumerate(c_list):  # Iterating over all labels in multiclass case
            label_value = self.explainer.check_label_name(label)[2] if multiclass else ''
            fig_features_importance = self.explainer.plot.features_importance(label=label)

            explain_contrib_data = list()
            list_cols_labels = [self.explainer.features_dict.get(col, col)
                                for col in self.col_names]
            for feature_label in sorted(list_cols_labels):
                feature = self.explainer.inv_features_dict.get(feature_label, feature_label)
                fig = self.explainer.plot.contribution_plot(feature, label=label, max_points=200)
                explain_contrib_data.append({
                    'feature_index': int(inv_columns_dict[feature]),
                    'name': feature,
                    'description': self.explainer.features_dict[feature],
                    'plot': plotly.io.to_html(fig, include_plotlyjs=False, full_html=False)
                })
            explain_data.append({
                'index': index_label,
                'name': label_value,
                'feature_importance_plot': plotly.io.to_html(fig_features_importance, include_plotlyjs=False,
                                                             full_html=False),
                'features': explain_contrib_data
            })
        print_html(explainability_template.render(labels=explain_data))
        print_md('---')

    def display_model_performance(self):
        """
        Displays the performance of the model. The metrics are computed using the config dict.

        Metrics should be given as a list of dict. Each dict contains they following keys :
        'path' (path to the metric function, ex: 'sklearn.metrics.mean_absolute_error'),
        'name' (optional, name of the metric as displayed in the report),
        and 'use_proba_values' (optional, possible values are False (default) or True
        if the metric uses proba values instead of predicted values).

        For example :
        config['metrics'] = [
                {
                    'path': 'sklearn.metrics.mean_squared_error',
                    'name': 'Mean absolute error',  # Optional : name that will be displayed next to the metric
                    'y_pred': 'predicted_values'  # Optional
                },
                {
                    'path': 'Scoring_AP.utils.lift10',  # Custom function path
                    'name': 'Lift10',
                    'y_pred': 'proba_values'  # Use proba values instead of predicted values
                }
            ]
        """
        if self.y_test is None:
            logging.info("No labels given for test set. Skipping model performance part")
            return

        print_md("### Univariate analysis of target variable")
        df = pd.concat([pd.DataFrame({self.target_name: self.y_pred}).assign(_dataset="pred"),
                        pd.DataFrame({self.target_name: self.y_test}).assign(_dataset="true")
                        if self.y_test is not None else None]).reset_index(drop=True)
        self._perform_and_display_analysis_univariate(
            df=df,
            col_splitter="_dataset",
            split_values=["pred", "true"],
            names=["Prediction values", "True values"],
            group_id='target-distribution'
        )

        if 'metrics' not in self.config.keys():
            logging.info("No 'metrics' key found in report config dict. Skipping model performance part.")
            return
        print_md("### Metrics")

        for metric in self.config['metrics']:
            if 'name' not in metric.keys():
                metric['name'] = metric['path']

            if metric['path'] in ['confusion_matrix', 'sklearn.metrics.confusion_matrix'] or \
                    metric['name'] == 'confusion_matrix':
                print_md(f"**{metric['name']} :**")
                print_html(convert_fig_to_html(generate_confusion_matrix_plot(
                    y_true=self.y_test,
                    y_pred=self.y_pred,
                    colors_dict=self.explainer.colors_dict
                )))
            else:
                try:
                    metric_fn = get_callable(path=metric['path'])
                    #  Look if we should use proba values instead of predicted values
                    if 'use_proba_values' in metric.keys() and metric['use_proba_values'] is True:
                        y_pred = self.explainer.proba_values
                    else:
                        y_pred = self.y_pred
                    res = metric_fn(self.y_test, y_pred)
                except Exception as e:
                    logging.info(f"Could not compute following metric : {metric['path']}. \n{e}")
                    continue
                if isinstance(res, Number):
                    res = display_value(round_to_k(res, 3))
                    print_md(f"**{metric['name']} :** {res}")
                elif isinstance(res, (list, tuple, np.ndarray)):
                    print_md(f"**{metric['name']} :**")
                    print_html(pd.DataFrame(res).to_html(classes="greyGridTable"))
                elif isinstance(res, str):
                    print_md(f"**{metric['name']} :**")
                    print_html(f"<pre>{res}</pre>")
                else:
                    logging.info(f"Could not compute following metric : {metric['path']}. \n"
                                 f"Result of type {res} cannot be displayed")
        print_md('---')



"""
sklearn columntransformer
"""

import pandas as pd
import numpy as np
from shapash.utils.category_encoder_backend import inv_transform_ordinal
from shapash.utils.category_encoder_backend import inv_transform_ce
from shapash.utils.category_encoder_backend import supported_category_encoder
from shapash.utils.category_encoder_backend import dummies_category_encoder
from shapash.utils.category_encoder_backend import category_encoder_binary
from shapash.utils.category_encoder_backend import transform_ordinal, get_col_mapping_ce
from shapash.utils.model_synoptic import simple_tree_model_sklearn, catboost_model,\
    linear_model, svm_model, xgboost_model, lightgbm_model, dict_model_feature
from shapash.utils.model import extract_features_model

columntransformer = "<class 'sklearn.compose._column_transformer.ColumnTransformer'>"

sklearn_onehot = "<class 'sklearn.preprocessing._encoders.OneHotEncoder'>"
sklearn_ordinal = "<class 'sklearn.preprocessing._encoders.OrdinalEncoder'>"
sklearn_standardscaler = "<class 'sklearn.preprocessing._data.StandardScaler'>"
sklearn_quantiletransformer = "<class 'sklearn.preprocessing._data.QuantileTransformer'>"
sklearn_powertransformer = "<class 'sklearn.preprocessing._data.PowerTransformer'>"

sklearn_model = linear_model + svm_model + simple_tree_model_sklearn

other_model = xgboost_model + catboost_model + lightgbm_model

dummies_sklearn = (sklearn_onehot)

no_dummies_sklearn = (sklearn_ordinal,
                      sklearn_standardscaler,
                      sklearn_quantiletransformer,
                      sklearn_powertransformer)

supported_sklearn = (sklearn_onehot,
                     sklearn_ordinal,
                     sklearn_standardscaler,
                     sklearn_quantiletransformer,
                     sklearn_powertransformer)


def inv_transform_ct(x_in, encoding):
    """
    Inverse transform when using a ColumnsTransformer.

    As ColumnsTransformer output hstack the result of transformers, if the TOP-preprocessed data are re-ordered
    after the ColumnTransformer the inverse transform must return false result.

    We successively inverse the transformers with columns position. That's why inverse colnames
    are prefixed by the transformers names.

    Parameters
    ----------
    x_in : pandas.DataFrame
        Prediction set.
    encoding : list
        The list must contain a single ColumnsTransformer and an optional list of dict.

    Returns
    -------
    pandas.Dataframe
        The reversed transformation for the given list of encoding.
    """
    if str(type(encoding)) == columntransformer:
        # We use inverse tranform from the encoding method base on columns position
        init = 0
        rst = pd.DataFrame()

        for enc in encoding.transformers_:
            name_encoding = enc[0]
            ct_encoding = enc[1]
            col_encoding = enc[2]
            # For Scikit encoding we use the associated inverse transform method
            if str(type(ct_encoding)) in supported_sklearn:
                frame, init = inv_transform_sklearn_in_ct(x_in,
                                                          init,
                                                          name_encoding,
                                                          col_encoding,
                                                          ct_encoding)

            # For category encoding we use the mapping
            elif str(type(ct_encoding)) in supported_category_encoder:
                frame, init = inv_transform_ce_in_ct(x_in,
                                                     init,
                                                     name_encoding,
                                                     col_encoding,
                                                     ct_encoding)

            # columns not encode
            elif name_encoding == 'remainder':
                if ct_encoding == 'passthrough':
                    nb_col = len(col_encoding)
                    frame = x_in.iloc[:, init:init + nb_col]
                else:
                    frame = pd.DataFrame()

            else:
                raise Exception(f'{ct_encoding} is not supported yet.')

            rst = pd.concat([rst, frame], axis=1)

    elif str(type(encoding)) == "<class 'list'>":
        rst = inv_transform_ordinal(x_in, encoding)

    else:
        raise Exception(f"{encoding.__class__.__name__} not supported, no inverse done.")

    return rst


def inv_transform_ce_in_ct(x_in, init, name_encoding, col_encoding, ct_encoding):
    """
    Inverse transform when using category_encoder in ColumnsTransformer preprocessing.

    Parameters
    ----------
    x_in : pandas.DataFrame
        Data processed.
    init : np.int
        Columns index that give the first column to look at.
    name_encoding : String
        Name of the encoding give by the user.
    col_encoding : list
        Processed features name.
    ct_encoding : category_encoder
        Type of encoding.

    Returns
    -------
    frame : pandas.Dataframe
        The reversed transformation for the given list of encoding.
    init : np.int
        Index of the last column use to make the transformation.
    """
    colname_output = [name_encoding + '_' + val for val in col_encoding]
    colname_input = ct_encoding.get_feature_names()
    nb_col = len(colname_input)
    x_to_inverse = x_in.iloc[:, init:init + nb_col].copy()
    x_to_inverse.columns = colname_input
    frame = inv_transform_ce(x_to_inverse, ct_encoding)
    frame.columns = colname_output
    init += nb_col
    return frame, init


def inv_transform_sklearn_in_ct(x_in, init, name_encoding, col_encoding, ct_encoding):
    """
    Inverse transform when using sklearn in ColumnsTransformer preprocessing.

    Parameters
    ----------
    x_in : pandas.DataFrame
        Data processed.
    init : np.int
        Columns index that give the first column to look at.
    name_encoding : String
        Name of the encoding give by the user.
    col_encoding : list
        Processed features name.
    ct_encoding : sklearn, category_encoder
        Type of encoding.

    Returns
    -------
    frame : pandas.Dataframe
        The reversed transformation for the given list of encoding.
    init : np.int
        Index of the last column use to make the transformation.
    """
    colname_output = [name_encoding + '_' + val for val in col_encoding]
    if str(type(ct_encoding)) in dummies_sklearn:
        colname_input = ct_encoding.get_feature_names(col_encoding)
        nb_col = len(colname_input)
    else:
        nb_col = len(colname_output)
    x_inverse = ct_encoding.inverse_transform(x_in.iloc[:, init:init + nb_col])
    frame = pd.DataFrame(x_inverse, columns=colname_output, index=x_in.index)
    init += nb_col
    return frame, init


def calc_inv_contrib_ct(x_contrib, encoding, agg_columns):
    """
    Reversed contribution when ColumnTransformer is used.

    As columns transformers output hstack the result of transformers, if the TOP-preprocessed data are re-ordered
    after the ColumnTransformer the inverse transform must return false result.

    Parameters
    ----------
    x_contrib : pandas.DataFrame
        Contributions set.
    encoding : ColumnTransformer, list, dict
        The processing apply to the original data.
    agg_columns : str (default: 'sum')
        Type of aggregation performed. For Shap we want so sum contributions of one hot encoded variables.
        For ACV we want to take any value as ACV computes contributions of coalition of variables (like
        one hot encoded variables) differently from Shap and then give the same value to each variable of the
        coalition. As a result we just need to take the value of one of these variables to get the contribution
        value of the group.

    Returns
    -------
    pandas.Dataframe
        The aggregate contributions depending on which processing is apply.
    """

    if str(type(encoding)) == columntransformer:
        # We use inverse tranform from the encoding method base on columns position
        init = 0
        rst = pd.DataFrame()

        for enc in encoding.transformers_:
            name_encoding = enc[0]
            ct_encoding = enc[1]
            col_encoding = enc[2]

            if str(type(ct_encoding)) in supported_category_encoder+supported_sklearn:
                # We create new columns names depending on the name of the transformers and the name of the column.
                colname_output = [name_encoding + '_' + val for val in col_encoding]

                # If the processing create multiple columns we find the number of original categories and aggregate
                # the contribution.
                if str(type(ct_encoding)) in dummies_sklearn or str(type(ct_encoding)) in dummies_category_encoder:
                    for i_enc in range(len(colname_output)):
                        if str(type(ct_encoding)) == sklearn_onehot:
                            col_origin = ct_encoding.categories_[i_enc]
                        elif str(type(ct_encoding)) == category_encoder_binary:
                            try:
                                col_origin = ct_encoding.base_n_encoder.mapping[i_enc].get('mapping').columns.tolist()
                            except:
                                col_origin = ct_encoding.mapping[i_enc].get('mapping').columns.tolist()
                        else:
                            col_origin = ct_encoding.mapping[i_enc].get('mapping').columns.tolist()
                        nb_col = len(col_origin)
                        if agg_columns == 'first':
                            contrib_inverse = x_contrib.iloc[:, init]
                        else:
                            contrib_inverse = x_contrib.iloc[:, init:init + nb_col].sum(axis=1)
                        frame = pd.DataFrame(contrib_inverse,
                                             columns=[colname_output[i_enc]],
                                             index=contrib_inverse.index)
                        rst = pd.concat([rst, frame], axis=1)
                        init += nb_col
                else:
                    nb_col = len(colname_output)
                    frame = x_contrib.iloc[:, init:init + nb_col]
                    frame.columns = colname_output
                    rst = pd.concat([rst, frame], axis=1)
                    init += nb_col

            elif name_encoding == 'remainder':
                if ct_encoding == 'passthrough':
                    nb_col = len(col_encoding)
                    frame = x_contrib.iloc[:, init:init + nb_col]
                    rst = pd.concat([rst, frame], axis=1)
                    init += nb_col
            else:
                raise Exception(f"{encoding.__class__.__name__} not supported, no inverse done.")
        return rst
    else:
        return x_contrib


def transform_ct(x_in, model, encoding):
    """
    Transform when using a ColumnsTransformer.

    As ColumnsTransformer output hstack the result of transformers, if the TOP-preprocessed data are re-ordered
    after the ColumnTransformer the inverse transform must return false result.

    We successively apply the transformers with columns position. That's why colnames
    are prefixed by the transformers names.

    Parameters
    ----------
    x_in : pandas.DataFrame
        Raw dataset to apply preprocessing
    model: model object
        model used to check the different values of target estimate predict_proba
    encoding : list
        The list must contain a single ColumnsTransformer and an optional list of dict.

    Returns
    -------
    pandas.Dataframe
        The data preprocessed for the given list of encoding.
    """
    if str(type(encoding)) == columntransformer:
        # We use inverse tranform from the encoding method base on columns position
        if str(type(model)) in sklearn_model:
            rst = pd.DataFrame(encoding.transform(x_in),
                               index=x_in.index)
            rst.columns = ["col_" + str(feature) for feature in rst.columns]

        elif str(type(model)) in other_model:
            rst = pd.DataFrame(encoding.transform(x_in),
                               columns=extract_features_model(model, dict_model_feature[str(type(model))]),
                               index=x_in.index)
        else:
            raise ValueError("Model specified isn't supported by Shapash.")

    elif str(type(encoding)) == "<class 'list'>":
        rst = transform_ordinal(x_in, encoding)

    else:
        raise Exception(f"{encoding.__class__.__name__} not supported, no preprocessing done.")

    return rst


def get_names(name, trans, column, column_transformer):
    """
    Allow to extract features names from one encoder of the ColumnTransformer.
    If the right names aren't available, It creates a list with customized names.

    Parameters
    ----------
    name: string
        String which indicates the name of the transformer.
    trans: sklearn encoder or category_encoders
        One of the encoder fitted through the ColumnTransformer.
    column: list
        List of features impacted by the specific transformer.
    column_transformer: sklearn ColumnTransformer
        The fitted ColumnTransformer containing the specific encoder.

    Returns
    -------
    list:
        List of returned features when specific transformer is applied.
    """
    if trans == 'drop' or (
            hasattr(column, '__len__') and not len(column)):
        return []
    if trans == 'passthrough':
        if hasattr(column_transformer, '_df_columns'):
            if ((not isinstance(column, slice))
                    and all(isinstance(col, str) for col in column)):
                return column
            else:
                return column_transformer._df_columns[column]
        else:
            indices = np.arange(column_transformer._n_features)
            return ['x%d' % i for i in indices[column]]
    if not hasattr(trans, 'get_feature_names'):
        if column is None:
            return []
        else:
            return [name + "__" + f for f in column]

    return [name + "__" + f for f in trans.get_feature_names()]


def get_feature_names(column_transformer):
    """
    Allow to extract all features names from encoders of the ColumnTransformer once it has been applied.
    If the right names aren't available, It creates a list with customized names.

    Parameters
    ----------
    column_transformer: sklearn ColumnTransformer
        The fitted ColumnTransformer containing the specific encoder.

    Returns
    -------
    feature_names: list
        List of returned features names when ColumnTransformer is applied.
    """
    feature_names = []
    l_transformers = list(column_transformer._iter(fitted=True))

    for name, trans, column, _ in l_transformers:
        feature_names.extend(get_names(name, trans, column, column_transformer))

    return feature_names


def get_list_features_names(list_preprocessing, columns_dict):
    """
    Allow to extract all features names from encoders when a list of preprocessing is uesd once it has been applied.
    If the right names aren't available, It creates a list with customized names.

    Parameters
    ----------
    list_preprocessing: list
        The fitted list_preprocessing containing the specific encoders.

    Returns
    -------
    feature_names: list
        List of returned features names when list_preprocessing is applied.

    """
    feature_expected = [value for value in columns_dict.values()]

    for enc in list_preprocessing:
        if str(type(enc)) in supported_category_encoder:
            feature_expected = enc.feature_names

        elif str(type(enc)) in columntransformer:
            feature_expected = get_feature_names(enc)

        elif str(type(enc)) in ("<class 'list'>"):
            feature_expected = feature_expected

    return feature_expected


def get_feature_out(estimator, feature_in):
    """
    Returns estimator features out if it has get_feature_names method, else features_in
    """
    if hasattr(estimator, 'get_feature_names'):
        return estimator.get_feature_names(), estimator.categories_
    else:
        return feature_in, []


def get_col_mapping_ct(encoder, x_encoded):
    """
    Get the columns mapping of a column transformer encoder.

    Parameters
    ----------
    encoder : ColumnTransformer
        The encoder used.
    x_encoded : pd.DataFrame
        Pandas dataframe after encoder transformations

    Returns
    -------
    dict_col_mapping : dict
        Dict of mapping between dataframe columns before and after encoding.
    """
    dict_col_mapping = dict()
    idx_encoded = 0
    for name, estimator, features in encoder.transformers_:
        if name != 'remainder':

            if str(type(estimator)) in dummies_sklearn:
                features_out, categories_out = get_feature_out(estimator, features)
                for i, f_name in enumerate(features):
                    dict_col_mapping[name + '_' + f_name] = list()
                    for _ in categories_out[i]:
                        dict_col_mapping[name + '_' + f_name].append(x_encoded.columns.to_list()[idx_encoded])
                        idx_encoded += 1

            elif str(type(estimator)) in no_dummies_sklearn:
                features_out, categories_out = get_feature_out(estimator, features)
                for f_name in features_out:
                    dict_col_mapping[name + '_' + f_name] = [x_encoded.columns.to_list()[idx_encoded]]
                    idx_encoded += 1

            elif str(type(estimator)) in supported_category_encoder:
                dict_mapping_ce = get_col_mapping_ce(estimator)
                for f_name in dict_mapping_ce.keys():
                    dict_col_mapping[name + '_' + f_name] = list()
                    for _ in dict_mapping_ce[f_name]:
                        dict_col_mapping[name + '_' + f_name].append(x_encoded.columns.to_list()[idx_encoded])
                        idx_encoded += 1

            else:
                raise NotImplementedError(f'Estimator not supported : {estimator}')

        elif estimator == 'passthrough':
            try:
                features_out = encoder.feature_names_in_[features]
            except:
                features_out = encoder._feature_names_in[features] #for oldest sklearn version
            for f_name in features_out:
                dict_col_mapping[f_name] = [x_encoded.columns.to_list()[idx_encoded]]
                idx_encoded += 1

    return dict_col_mapping



"""
load_smartpredictor module
"""
from shapash.explainer.smart_predictor import SmartPredictor
from shapash.utils.io import load_pickle


def load_smartpredictor(path):
    """
    load_smartpredictor allows Shapash users to load SmartPredictor Object already saved into a pickle.

    Parameters
    ----------
    path : str
        File path of the pickle file.

    Example
    --------
    >>> predictor = load_smartpredictor('path_to_pkl/predictor.pkl')
    """
    predictor = load_pickle(path)
    if isinstance(predictor, SmartPredictor):
        return predictor
    else:
        raise ValueError(
            f"{predictor} is not an instance of type SmartPredictor"
        )
